/*
data structures which describe registers on the Xilinx
Complex Programmable Logic Device (CPLD) attached between
the AMCC PCI bus interface and the RX FPGA, TX FPGA, and
AIM optics module.



Note 1: these data structures use bit fields, which are compiler-
and Central Processing Unit (CPU) specific, so you will probably
have to rewrite these data structures if you port to another
CPU or compiler

On the Intel CPU's using the Borland C++ 4.5 compiler, bits are
assigned from least significant to most significant within a word,
and less significant bytes of a word are stored at lower memory
addresses

No padding is inserted, and bitfields can straddle word boundaries

The width of each bitfield must be from 1 to 16 (see page 62 of the
Borland C++ version 4.5 Programmer's Guide).



Note 2: due to the use of anonymous unions, this header can only be
included from C++ code, not C code



Note 3: definitions of readability/writability abbreviations:
    W/O     writable only
    R/O     readable only
    W/R     writable and readable
    CW/R    readable, but writable only after
            CWRE bit set in Control_register
*/



#ifndef _AIM_FPGA_H
#define _AIM_FPGA_H



#include "../types/types.h"
#include "../apptel/xfpga.h"



struct AimModule1_register
{
    Bit8 DeviceType     : 4;
    Bit8 Reserved       : 4;
};



enum AimLineInterface
{
    MultiMode,
    SingleMode
};

struct AimModule2_register
{
    Bit8 RevisionAndCustomizationLevel      : 4;
    Bit8 LineInterfaceType                  : 4;
};



enum AimClockSelect
{
    InternalOscillator,
    ExternalTimingReference,
    InternalOscillatorAlias,
    ReceivedClock
};

enum AimClockMode
{
    Forced,                 // always as specified by clock select
    Revertive,              // sel. clock until failure, then int., go back when sel. comes back
    NonRevertive            // sel. clock until failure, then stay with int. even if sel. comes back
};

struct AimControl1_register
{
    Bit8 ClockSelect                    : 2;
    Bit8 ClockMode                      : 2;
    Bit8 InternalOscillatorIsActive     : 1;
    Bit8 Reserved                       : 3;
};



struct AimStatus1_register
{
    Bit8 LostExternalClockTransition    : 1;    // latched until read
    Bit8 LostReceivedDataTransition     : 1;    // latched until read
    Bit8 Reserved1                      : 2;
    Bit8 LostExternalClock              : 1;
    Bit8 LostReceivedData               : 1;
    Bit8 Reserved2                      : 2;

};



struct AimControl2_register
{
    // when cleared, (power on and FPGA configured) drives active LED
    // when set, software controls the active LED
    //
    Bit8 EnableActiveLedControl         : 1;

    // set by software when active LED should be on
    //
    // ???never affected by hardware
    //
    Bit8 ActiveLedState                 : 1;

    // when cleared, (FPGA not configured) drives fault LED
    // when set, software controls the fault LED
    //
    Bit8 EnableFaultLedControl          : 1;

    // set by software when fault LED should be on
    //
    // ???never affected by hardware
    //
    Bit8 FaultLedState                  : 1;

    // when cleared, (LOS or LOF) drives active LED
    // when set, software controls the active LED
    //
    Bit8 EnableLineLedControl           : 1;

    // set by software when line LED should be on
    //
    // ???never affected by hardware
    //
    Bit8 LineLedState                   : 1;

    Bit8 Reserved                       : 2;
};



// note: since the *INT line from the AIM-622 is not wired
// to anything on the POINT (PCI card), this register is
// irrelevant to OC12MON
//
struct AimInterruptEnable_register
{
    Bit8 LostExternalClockTransition    : 1;
    Bit8 LostReceivedDataTransition     : 1;
    Bit8 Reserved                       : 6;
};



enum AimOverhead0Selection
{
    Sts3Number1,
    Sts3Number2,
    Sts3Number3,
    Sts3Number4,
};

struct AimOverheadPortControl1_register
{
    Bit8 Port0Select    : 2;
    Bit8 Reserved       : 6;
};



enum AimOverhead2Or3Select
{
    NoOverhead,
    SectionDDC,
    SectionOrderwire,
    SectionUserChannel,
    LineDDC,
    LineOrderwire,
    LineAutomaticProtectSwitching
};

struct AimOverheadPortControl2_register
{
    Bit8 Port2Select    : 4;
    Bit8 Port3Select    : 4;
};



enum LoopbackSelection
{
    NoLoopback,
    Equipment,                      // both world and me RX what I TX
    Facility,                       // both world and me RX what world TX
    Split                           // world RX what world TX, I RX what I TX
};

struct AimLoopbackControl_register
{
    Bit8 Select     : 2;
    Bit8 Reserved   : 6;
};



// merged registers
//
struct Aim_registers
{
    union
    {
        AimModule1_register Module1;
        Bit32 dword_0;
    };

    union
    {
        AimModule2_register Module2;
        Bit32 dword_1;
    };

    union
    {
        AimControl1_register Control1;
        Bit32 dword_2;
    };

    union
    {
        AimStatus1_register Status1;
        Bit32 dword_3;
    };

    union
    {
        AimControl2_register Control2;
        Bit32 dword_4;
    };

    union
    {
        AimInterruptEnable_register InterruptEnable;
        Bit32 dword_5;
    };

    union
    {
        AimOverheadPortControl1_register OverheadPortControl1;
        Bit32 dword_6;
    };

    union
    {
        AimOverheadPortControl2_register OverheadPortControl2;
        Bit32 dword_7;
    };

    union
    {
        AimLoopbackControl_register LoopbackControl;
        Bit32 dword_8;
    };

    Bit32 Reserved1[ 15 ];

    union
    {
        AddressXor_register AddressXor;      // looks at 5 address LSB's
        Bit32 dword_24;
    };

    union
    {
        Diagnostic_register Diagnostic;
        Bit32 dword_25;
    };

    Bit32 AimReserved2[ 230 ];
};



#endif      // not included yet
