#ifndef _FORE_DRIVER_H
#define _FORE_DRIVER_H



// headers we depend upon
//
#include "../intel/i8259a.h"            // class InterruptController
#include "../fore/pca200e.h"
#include "../integrat/constant.h"       // whether cell payload got copied to host and for which cells
#include "../integrat/buffer.h"         // class BufferBlock class BufferQueue



// conditional compile switches that our clients use
//
#define CARD_INIT_PROVIDES_BUFFERS      FALSE
#define ATM_HANDLER_BYPASSED            FALSE



// error return codes
//
#define ERROR_CARD_BASE                 2000
#define ERROR_CARD_SHIFT_TOO_SLOW       (ERROR_CARD_BASE + 0)
#define ERROR_CARD_CANNOT_OPEN_FILE     (ERROR_CARD_BASE + 1)
#define ERROR_CARD_CANNOT_READ_HDR      (ERROR_CARD_BASE + 2)
#define ERROR_CARD_CANNOT_READ_BODY     (ERROR_CARD_BASE + 3)
#define ERROR_CARD_CANNOT_SEEK_END      (ERROR_CARD_BASE + 4)
#define ERROR_CARD_CANNOT_SEEK_BEG      (ERROR_CARD_BASE + 5)
#define ERROR_CARD_CANNOT_TELL_SIZE     (ERROR_CARD_BASE + 6)
#define ERROR_CARD_BSTAT_STUCK          (ERROR_CARD_BASE + 7)
#define ERROR_CARD_CP_NOT_RUNNING       (ERROR_CARD_BASE + 8)
#define ERROR_CARD_SELF_TEST_FAILED     (ERROR_CARD_BASE + 9)
#define ERROR_CARD_SELF_TEST_UNKNOWN    (ERROR_CARD_BASE + 10)
#define ERROR_CARD_HEARTBEAT_STUCK      (ERROR_CARD_BASE + 11)
#define ERROR_CARD_CP_FAULTED           (ERROR_CARD_BASE + 12)
#define ERROR_CARD_CMD_STUCK            (ERROR_CARD_BASE + 13)
#define ERROR_CARD_CMD_INVALID          (ERROR_CARD_BASE + 14)
#define ERROR_CARD_CMD_FAILED           (ERROR_CARD_BASE + 15)
#define ERROR_CARD_PREV_CMD_PENDING     (ERROR_CARD_BASE + 16)
#define ERROR_CARD_NOT_FOUND            (ERROR_CARD_BASE + 17)
#define ERROR_CARD_UNMAPPED             (ERROR_CARD_BASE + 18)
#define ERROR_CARD_CANNOT_MAP_PHYS      (ERROR_CARD_BASE + 19)



// manifest constants
//
#define FIRMWARE_PATH           "\\CORAL\\BIN\\"
#if !CARD_COPIES_ATM_PAYLOAD
    #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "only_atm.bin"
#else
#if CARD_COPIES_ALL_CELLS
    #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "pay_all.bin"
#else
#if CARD_SKIPS_GIGASWITCH
    #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "skipgigx.bin"
#else
#if CELLS_PER_PACKET == 0
#if CARD_COPIES_LAST_CELL
    #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "1st_last.bin"
#endif
#else
#if CELLS_PER_PACKET == 1
#if CARD_COPIES_LAST_CELL
    #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "1st_last.bin"
#else
    #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "pca200e.bin"
#endif
#else
#if CELLS_PER_PACKET == 2
    #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "2cell.bin"
#else
#if CELLS_PER_PACKET == 3
    #define FIRMWARE_FILE_NAME  FIRMWARE_PATH "3cell.bin"
#endif
#endif
#endif
#endif
#endif
#endif
#endif






class AtmCard
{

public:
    int init_result;

    Bit8 card_index;
    Bit16 card_handle;

    Bit32 phys_handle;
    Bit32 phys_address;
    pca200e_memmap *memmap;

    // constructor
    //
    AtmCard( Bit8 cardindex );

    // destructor
    //
    ~AtmCard();
};



enum LoopbackSelection
{
    NoLoopback,
    Facility                        // both world and me RX what world TX
};

// all the data that is per-card
//
class AtmCardRunning : public AtmCard
{

public:
    static InterruptController irq;
    unsigned irq_level;
    unsigned char interrupt_number;
    vfp thunk;
    Bit48 old_handler;
    enum InitType
    {
        NormalAtm,
        NormalAtmWith16BitTimestamps,
        PassIdleCells
    };


    // this field is public so it can be passed to the
    // foreground stats code when a remote query occurs,
    // so the query will reflect all data seen so far,
    // in case it will be 10 days until the block fills
    //
    // maybe we should use an accessor function instead?
    //
    BufferBlock *my_block;

private:
    enum AtmShiftTiming
    {
        during_initialization,
        during_interrupt,
        during_cleanup
    };

    Bit8 old_pci_latency_timer;
    CardClock atm_interrupt_time;
    PentiumClock atm_enabled;
    PentiumClock before;
    PentiumClock elapsed;
    PentiumClock elapsed_tot;
    PentiumClock elapsed_min;
    PentiumClock elapsed_max;
    int atm_interrupt_error;
    int timer_interrupts_allowed;
    unsigned atm_out_fifo_full;
    unsigned atm_in_fifo_almost_full;
    unsigned long atm_interrupt_count;
    unsigned long atm_interrupt_unexpected;
    unsigned long sonet_interrupt_count;
    unsigned long timer_interrupt_count;
    union { HostControl_read atm_status; Bit32 atm_status_alias; };
    union { PciMasterControl old_pci_master_control; Bit8 old_pci_master_control_alias; };
    int atm_uart_new_line;

    // note: my_block has been moved to public area, below
    BufferBlock *next_block;
    BufferBlock *my_tx_block;
    BufferBlock *next_tx_block;

    Status *atm_command_result;
    Qcard_cmd *atm_command;
    const char *atm_command_name;
    PentiumClock atm_command_start_time;

#if ATM_SHIFT_CLEANUP_DEBUG
    Bit32 bufcount_overflow_first_used;
    Bit32 bufcount_overflow_last_used;
    Bit32 bufcount_overflow_highest_used;
    Bit32 bufcount_nonoverflow_last_used;
    Bit32 bufcount_overflow_sar_rx_free;
    unsigned bufcount_overflow_prev_free;
    unsigned bufcount_overflow_this_free;
    unsigned bufcount_overflow_next_free;
    Bit32 bufcount_overflow_phys_free_base;
    PentiumClock bufcount_overflow_first_time;
    PentiumClock bufcount_overflow_last_time;
    PentiumClock bufcount_overflow_highest_time;
    PentiumClock bufcount_nonoverflow_last_time;
#endif

    // booleans set during card_init()
    //
    int rx_allowed;
    int tx_allowed;

public:
    // gets the card running the way you want
    //
    int card_init( InitType itype, LoopbackSelection ltype, int want_rx, int want_tx );

    void tick( void );

    int get_interrupt_result( void )
    {
        int result = atm_interrupt_error;
        atm_interrupt_error = 0;
        return result;
    }

    // called when BufferQueue free list stops being empty
    // in case this card was in need of a buffer
    //
    void foreground_replenish( void );

    // called when BufferQueue sendable list stops being empty
    // for this card in particular, in case this card was
    // in need of a buffer
    //
    void foreground_tx_replenish( void );

    int start_clearing_timestamp( void );
    int wait_for_timestamp_to_clear( void );

    void set_pending_block_usage( void );
    void set_pending_tx_block_usage( void );

    void timer_handler( void );

    // constructor
    //
    AtmCardRunning( Bit8 card_index0 );

    // destructor
    //
    ~AtmCardRunning();

private:
    // static interrupt handler that we can take the address of,
    // so the thunk can call it
    //
    static int interrupt_handler0( AtmCardRunning *this0,
        Bit32 ebp, Bit32 edi, Bit32 esi,
        Bit32 ds, Bit32 es,
        Bit32 edx, Bit32 ecx, Bit32 ebx, Bit32 eax,
        Bit32 eip, Bit32 cs, Bit32 flags
    )
    {
#if ATM_HANDLER_BYPASSED
        flat_put_long( 0xFFFBE008, 0 );
        InterruptController::acknowledge( 3 );
        return 0;
#else
        return this0->interrupt_handler();
#endif
    }

    int interrupt_handler( void );
    void interrupt_init( void );
    void interrupt_exit( void );

    void sonet_init( void );
    void sonet_exit( void );
    void sonet_handler( void );

    void atm_uart_io( const char *output, char *input, unsigned input_size, double delay );

    int atm_reset_hardware( void );

    int atm_download( void );

    int atm_check_heartbeat( int want_increasing );

    int atm_await_command( double uart_delay, double command_delay, volatile Status *status_location );

    int atm_shutdown_command( void );

    int atm_init_command( InitType itype, LoopbackSelection ltype );

    int atm_init( InitType itype, LoopbackSelection ltype );
    void atm_exit( void );
    void atm_handler( void );

    int atm_rx_completed( AtmShiftTiming timing );
    int atm_tx_completed( AtmShiftTiming timing );
};






#endif      // not included yet
