/* conditional compile switches all go here
 */

#define BYTESWAP_FIELDS_EARLY   FALSE       /* must match fs2flows.c */
#define USE_OLD_HASH            FALSE
#define ALWAYS_DUMP_HASH_INFO   FALSE
#define FOLD_BEFORE_MODULUS     FALSE
#define FOLD_RANDOMIZES         TRUE
#define FOLD_KEEPS_16_BITS      TRUE
#define DEBUG_CHAIN_LENGTH      FALSE
#define DEBUG_HASHINDEX         FALSE
#define DEBUG_WORKING_SET       TRUE
#define DEBUG                   FALSE
#define SORT_FRAG_PRINTOUT      FALSE
#define TALK_WHEN_ALLOC_FAILS   FALSE           // for debugging when overflow occurs only!  too slow
#define ENABLE_THRESHOLDS       FALSE
#define REPORT_WITH_RATIOS      FALSE




#if __unix__
#   include <sys/types.h>
#   include <sys/time.h>
#   include <netinet/in.h>
#else
#   include "../wattcp/tcp.h"
#endif
#include <stdio.h>
#include <stdlib.h>             /* errno */
#include <fcntl.h>

#if __unix__
#   include "types.h"
#else
#   include "../types/types.h"
#   include "../intel/byteswap.h"
#endif

#include "queue.h"
#include "trace.h"              /* must be included before util.h */
#include "util.h"
#include "dostats.h"            /* nonzero() variance() */
#include "fragstat.h"
#include "../dpmi32/flat.h"     /* flat_alloc() */



// none of this code gets compiled in
// if it's isn't going to be called
//
#if DO_FOREGROUND_STATS && REASM_IP_FRAGS



static int      *frag_ref = NULL;
static int      fragthreshold;
static float    cutoffp, cutoffb, cutofff;
static int      fragpaircount;

#define N_FRAGBUCKETS    256

/* generate an 8-bit hash key
 */
#define fraghash(iface,prot,dport) (iface ^ prot ^ (dport >> 8) ^ (dport & 255))

static struct fraghashrec **fragtab = NULL;
#define FRAGBUCKET_POOL_SIZE (N_FRAGBUCKETS * sizeof(struct fraghashrec *))

static struct fraghashrec *fraghash_free = NULL;
static struct fraghashrec *fraghash_pool = NULL;
u_int fraghash_alloc_failures;

#define FRAGHASH_RECORDS 1000L              // ~1K for vBNS, ~100K for BIPP

#define FRAGHASH_POOL_SIZE (FRAGHASH_RECORDS * sizeof(struct fraghashrec))

#if DEBUG_CHAIN_LENGTH
#   define FRAGCHAIN_POOL_SIZE (N_FRAGBUCKETS * sizeof(int))
    static int *chainlen2 = NULL;
    static int longest_chain2;
#endif

#if DEBUG_WORKING_SET
    static int total_level2_fraghashrecs;
#endif

#if DEBUG_HASHINDEX
    static u_int hashindex_trace = -1;
#endif



#if SORT_FRAG_PRINTOUT
#define FRAGPRINT_POOL_SIZE (FRAGHASH_RECORDS * sizeof(struct fraghashrec *))
static Bit32 fragcount;
static struct fraghashrec **fragprintout = NULL;

static int fragcmp( void *a, void *b )
{
    struct fraghashrec ** hrp1 = (struct fraghashrec**) a;
    struct fraghashrec ** hrp2 = (struct fraghashrec**) b;
    return ((*hrp1)->packets - (*hrp2)->packets);
}
#endif



/*
 * new_fragment_record
 *      establish a new fragment pair
 *      1> insert into hash table [maintained as stack]
 */
static
struct fraghashrec *
new_fragment_record( Bit8 iface, u_int prot, u_int dport, u_int dst
#if KEEP_ATM_HEADER
    ,u_int atm
#endif
    ,u_int olaps, u_int holes
    ,u_int flows, u_int packets, Bit64 bytes, float duration
#if KEEP_FRAG_SQUARES
    ,double p2sum, double b2sum, double d2sum
    ,double pbsum, double bdsum, double pdsum
#endif
)
{
    struct fraghashrec *pr;

    if ((pr = fraghash_free) == NULL) {
#if TALK_WHEN_ALLOC_FAILS
        fprintf(stderr, "unable to alloc new fraghashrec\n");
#endif
        fraghash_alloc_failures++;
        return 0;
    }
    fraghash_free = pr->next;

#if KEEP_ATM_HEADER
    pr->n_atm=0;
#endif
    pr->iface=iface;
    pr->proto=prot;
    pr->dport=dport;
    pr->dst=dst;
#if KEEP_ATM_HEADER
    pr->atm[0].whole=atm;
#endif
    pr->olaps=olaps;
    pr->holes=holes;
    pr->flows=flows;
    pr->packets=packets;
    pr->bytes=bytes;
    pr->duration=duration;
#if KEEP_FRAG_SQUARES
    pr->p2sum=p2sum;
    pr->b2sum=b2sum;
    pr->d2sum=d2sum;
    pr->pbsum=pbsum;
    pr->bdsum=bdsum;
    pr->pdsum=pdsum;
#endif
    pr->next=NULL;
    return(pr);
}



static void delete_fragment_record( struct fraghashrec *pr )
{
    /* put this puppy on the free list */
    pr->next = fraghash_free;
    fraghash_free = pr;
}


/*
 * diagnostic
 */
static void print_fragment_record( struct fraghashrec *pr )
{
#if KEEP_ATM_HEADER
    unsigned count;
#endif

    fprintf(stderr, "packets: %d ", pr->packets);
    fprintf(stderr, "bytes: %.0f ", (double)pr->bytes);
    fprintf(stderr, "iface dst proto dport duration hash: %d %d %d %f %d\n",
        pr->iface,pr->dst,pr->proto,pr->dport,pr->duration,
        fraghash(pr->iface,pr->proto,pr->dport));
#if KEEP_ATM_HEADER
    fprintf( stderr, "  ATM" );
    for( count=0; count<=pr->n_atm && count!=MAX_FRAG_ATMS; count++ )
        fprintf( stderr, " %d/%d/%d",
            pr->atm[count].bits.GFC,
            pr->atm[count].bits.VPI,
            pr->atm[count].bits.VCI
        );
    if( pr->n_atm == MAX_FRAG_ATMS )
        fprintf( stderr, "+" );
    fprintf( stderr, "\n" );
#endif
}



// forgets all statistics
//
int init_fragment_stats( void )
{
    Bit32 count;

#if ENABLE_THRESHOLDS
    cutoffp = .1;           // 10%    of packets
    cutoffb = .0005;        //  0.05% of bytes
    cutofff = .1;           // 10%    of flows
#else
    cutoffp = 0.0;
    cutoffb = 0.0;
    cutofff = 0.0;
#endif

    fragpaircount = 0;

#if DEBUG_CHAIN_LENGTH
    longest_chain2 = 0;
#endif

#if DEBUG_WORKING_SET
    total_level2_fraghashrecs = 0;
#endif

    // if this is the first clear (really initialization)
    //
    if( fraghash_pool == NULL )
    {
        fprintf( stderr, "frag hash records are %d bytes each\n", sizeof(struct fraghashrec) );

#if DEBUG_CHAIN_LENGTH
        /* allocate chain length counters for each bucket in main hash table */
        chainlen2 = (int *) flat_alloc( FRAGCHAIN_POOL_SIZE );
        if( chainlen2 == NULL )
        {
            printf( "error %d allocating %d bytes (%d integers) for fragment chain length counters\n", errno, FRAGCHAIN_POOL_SIZE, N_FRAGBUCKETS );
            return 1;
        }
#endif

        /* allocate the buckets for the hash table */
        fragtab = (struct fraghashrec **) flat_alloc( FRAGBUCKET_POOL_SIZE );
        if( fragtab == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for fragment hash buckets\n", errno, FRAGBUCKET_POOL_SIZE, N_FRAGBUCKETS );
            return 1;
        }

#if SORT_FRAG_PRINTOUT
        /* allocate pointers to sort during report */
        fragprintout = (struct fraghashrec **) flat_alloc( FRAGPRINT_POOL_SIZE );
        if( fragprintout == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld pointers) for fragment printout sorting\n", errno, FRAGPRINT_POOL_SIZE, FRAGHASH_RECORDS );
            return 1;
        }
#endif

        /* allocate a pool of hash records */
        fraghash_pool = (struct fraghashrec *) flat_alloc( FRAGHASH_POOL_SIZE );
        if( fraghash_pool == NULL )
        {
            printf( "error %d allocating %ld bytes (%ld records) for fragment hash record pool\n", errno, FRAGHASH_POOL_SIZE, FRAGHASH_RECORDS );
            return 1;
        }

        /* put all the hash records on the singly-linked free list */
        fraghash_free = fraghash_pool;
        for( count=0; count<FRAGHASH_RECORDS-1; count++ )
        {
            fraghash_pool[ count ].next = fraghash_pool + count + 1;
        }
        fraghash_pool[count].next = NULL;

    }
    fraghash_alloc_failures = 0;

    for( count=0; count<N_FRAGBUCKETS; count++ )
    {
        // level2 hashtable is empty
        //
        fragtab[count] = 0;

#if SORT_FRAG_PRINTOUT
        // no frag pairs are ready to print
        //
        fragprintout[ count ] = 0;
#endif

#if DEBUG_CHAIN_LENGTH
        /* initialize chain length of all hashtable buckets to zero
         */
        chainlen2[ count ] = 0;
#endif
    }

    return 0;
}



void update_fragment_stats( struct hashrec *hr, u_char proto,
    u_int packets, Bit64 bytes, float duration
#if KEEP_FRAG_SQUARES
    ,double packets2, double bytes2, double duration2,
    double packets_bytes, double bytes_duration, double packets_duration
#endif
)
{
        int cmp;
        Bit8 fraghashindex;
        struct fraghashrec *pp, *fragh, *newpp;
        Bit8 iface = hr->hr_iface;
#if BYTESWAP_FIELDS_EARLY
        u_int  dport = hr->hr_dport;
        u_int    dst = hr->hr_dst;
#else
        u_int  dport = ntohs( hr->hr_dport );
        u_int    dst = ntohl( hr->hr_dst );
#endif
#if KEEP_ATM_HEADER
        Bit8 n_atm;
#endif

        // ignore the flow if it never had an overlap or hole
        //
        if( hr->hr_olaps==0 && hr->hr_holes==0 )
            return;

        fraghashindex = fraghash(iface,proto,dport);
        pp = fragtab[fraghashindex];
        /* look for match or collision on hash chain */
        fragh=pp;
        cmp=0;
        while (fragh!=NULL && cmp==0)
        {
                cmp = ( (iface==fragh->iface) && (proto==fragh->proto) &&
                        (dport==fragh->dport) && (dst==fragh->dst) );
                if (cmp) {
                        /* update all the stats in fraghashrec */
                        fragh->olaps+=hr->hr_olaps;
                        fragh->holes+=hr->hr_holes;
                        fragh->flows++;
                        fragh->packets+=packets;
                        fragh->bytes+=bytes;
                        fragh->duration+=duration;
#if KEEP_FRAG_SQUARES
                        fragh->p2sum+=packets2;
                        fragh->b2sum+=bytes2;
                        fragh->d2sum+=duration2;
                        fragh->pbsum+=packets_bytes;
                        fragh->bdsum+=bytes_duration;
                        fragh->pdsum+=packets_duration;
#endif
#if KEEP_ATM_HEADER
                        n_atm = fragh->n_atm;
                        if( n_atm != MAX_FRAG_ATMS )
                        {
                            n_atm++;
                            if( n_atm != MAX_FRAG_ATMS )
                                fragh->atm[ n_atm ].whole = hr->hr_atm.whole;

                            fragh->n_atm = n_atm;
                        }
#endif
#if DEBUG
                        /* for debugging */
                        printf("match\t");
#endif
                }
                fragh=fragh->next;
        };
        /* didn't find it anywhere, must be a new record */
        if (cmp==0)
        {
            fragpaircount++;
            fragh = new_fragment_record( iface, proto, dport, dst,
                hr->hr_atm.whole,
                hr->hr_olaps, hr->hr_holes, 1, packets, bytes, duration
#if KEEP_FRAG_SQUARES
                ,packets2, bytes2, duration2,
                packets_bytes, bytes_duration, packets_duration
#endif
            );
            if( fragh )
            {
                fragh->next = pp;
                fragtab[fraghashindex] = fragh;
#if DEBUG_WORKING_SET
                /* increment count of level2 hash records
                 */
                total_level2_fraghashrecs++;
#endif

#if DEBUG_CHAIN_LENGTH
                /*
                 * increment chain length for this hash index
                 */
                chainlen2[fraghashindex]++;

#if DEBUG_HASHINDEX
                if( fraghashindex == hashindex_trace )
                    fprintf( stderr, "at packet %d hash index 0x%8.8lX chain length increased to %d\n", total_packets, fraghashindex, chainlen2[fraghashindex] );
#endif

                /*
                 * if this is the longest chain we have ever seen,
                 * update our global, and notify the user
                 *
                 * notifications should be asymptotically decreasing
                 */
                if( longest_chain2 < chainlen2[fraghashindex] )
                {
                    longest_chain2 = chainlen2[fraghashindex];
                    fprintf( stderr, "packet %d hash index 0x%8.8lX is now longest fragment chain 2 at %d\n", total_packets, fraghashindex, longest_chain2 );
                    fprintf( stderr, "calced from proto=%d dport=%4.4X\n", proto, dport );
                }
#endif

/* for debugging */
#if DEBUG
                printf("new\t");
#endif
            }   // end if could alloc fraghashrec
        }   // end if needed a new fraghashrec
#if DEBUG
        print_fragment_record(fragh);
#endif
}



/* "SFB iface dst proto dport olaps holes "
   "flows packets bytes dur %%f %%p %%b %%d fps "
   "pps bps ppf bpf spf bpp vp vb vd copb cobd copd ATM1 ATM2 ATM3 "
   "ATM4 ATM5 ATM6 ATM7 ATM8 ATM9 ATM10 ATM11 ATM12 ATM13 ATM14 +\n");
*/
static void print_fragment_stats( tcp_Socket *sock
    ,struct fraghashrec *fragh
#if REPORT_WITH_RATIOS
    ,float trace_duration
    ,int total_flows
    ,double packetsum
    ,double bytesum
    ,double dursum
#endif
)
{
    char temp[20];
#if KEEP_ATM_HEADER
    unsigned count;
#endif

    sock_printf(
        sock,
        "SFB\t%d\t%s\t%d\t%d\t%d\t%d\t"
            "%d\t%d\t%.0f\t%.3f\t"
#if REPORT_WITH_RATIOS
            "%.3f\t%.3f\t%.3f\t%.3f\t"
            "%.3f\t%.0f\t%.0f\t"
            "%.3f\t%.0f\t%.3f\t%.0f\t"
#endif
#if KEEP_FRAG_SQUARES
            "%.3f\t%.3f\t%.3f\t"
            "%.3f\t%.3f\t%.3f"
#endif

        ,fragh->iface,
        inet_ntoa( temp, fragh->dst ),
        fragh->proto,
        fragh->dport,
        fragh->olaps,
        fragh->holes,

        fragh->flows,
        fragh->packets,
        (double) fragh->bytes,
        fragh->duration

#if REPORT_WITH_RATIOS
        ,(float) fragh->flows/total_flows,
        (float) fragh->packets/packetsum,
        (float) fragh->bytes/bytesum,
        (float) fragh->duration/dursum,

        fragh->flows/nonzero(trace_duration),
        fragh->packets/nonzero(trace_duration),
        fragh->bytes/nonzero(trace_duration),

        (float) fragh->packets/fragh->flows,
        (float) fragh->bytes/fragh->flows,
        (float) fragh->duration/fragh->flows,

        (float) fragh->bytes/fragh->packets
#endif

#if KEEP_FRAG_SQUARES
        ,variance(fragh->flows,fragh->packets,fragh->p2sum),
        variance(fragh->flows,fragh->bytes,fragh->b2sum),
        variance(fragh->flows,fragh->duration,fragh->d2sum),

        (float) fragh->pbsum,
        (float) fragh->bdsum,
        (float) fragh->pdsum                // 27th field
#endif
    );

#if KEEP_ATM_HEADER
    for( count=0; count<=fragh->n_atm && count!=MAX_FRAG_ATMS; count++ )
        sock_printf( sock, "\t%d/%d/%d",
            fragh->atm[count].bits.GFC,
            fragh->atm[count].bits.VPI,
            fragh->atm[count].bits.VCI
        );

    if( fragh->n_atm == MAX_FRAG_ATMS )
        sock_printf( sock, "\t+" );
#endif

    sock_printf( sock, "\n" );
}



// turn off compiler warnings about unused variables
// trace_duration and dursum
// because caller doesn't have the symbol defined
// to know whether we are supposed to get those parms
// because that is a local decision
//
#pragma warn -par

int report_fragment_stats( tcp_Socket *sock, float trace_duration,
    u_int flows, double packetsum, double bytesum, double dursum )
{
        Bit64 cutoffb_bit64;
        int i;
        int fraghashindex, cmp;
        struct fraghashrec *pp, *fragh, *next;

#if DEBUG_WORKING_SET
        fprintf( stderr, "%d level 2 frag hash records created\n", total_level2_fraghashrecs );
#endif

        sock_printf( sock, "SFC\t%d\n", fragpaircount );

        /* go through fragtab and print now */
        cutoffp *= packetsum;
        cutoffb *= bytesum;
        cutoffb_bit64 = cutoffb;
        cutofff *= flows;

        for (i=0;i<N_FRAGBUCKETS;i++) {
            for (fragh=fragtab[i]; fragh!=NULL; fragh=next) {
                next = fragh->next;

                if ((fragh->packets > cutoffp) ||
                    (fragh->bytes > cutoffb_bit64) ||
                    (fragh->flows > cutofff))
                {
#if SORT_FRAG_PRINTOUT
                    /* fragprintout: array of pointers to struct fraghashrec's */
                    fragcount++;
                    fragprintout[fragcount]=fragh;
#else
                    print_fragment_stats(sock, fragh
#if REPORT_WITH_RATIOS
                        ,trace_duration, flows, packetsum, bytesum, dursum
#endif
                    );
#endif
                }
#if SORT_FRAG_PRINTOUT
                /* only want to delete unselected frag pairs now;
                 * selected ones will be deleted after sort/print
                 */
                else
#else
                /* want to delete both printed and unprinted frag pairs
                 */
#endif
                    delete_fragment_record( fragh );
            }
        }

#if SORT_FRAG_PRINTOUT
        /* sort from fraghle array 80 each */
        qsort( fragprintout, fragcount, sizeof(struct fraghashrec *), &fragcmp );
        for( i=0; i<fragcount; i++)
        {
            print_fragment_stats( sock, fragprintout[i]
#if REPORT_WITH_RATIOS
                , trace_duration, packetsum, bytesum, dursum
#endif
            );
            delete_fragment_record( fragprintout[i] );
        }
#endif

    return 0;
}

// restore compiler warnings about unused variables
//
#pragma warn .par



#endif  // DO_FOREGROUND_STATS && REASM_IP_FRAGS
