#ifndef _I8259_H
#define _I8259_H



#include <dos.h>                        // enable() disable() inportb() outportb()

#ifdef __DPMI32__
// linker won't resolve these unless we get rid of the
// header file's attempt to use use inline functions,
// which the 32-bit compiler doesn't support
//
#undef outportb
#undef outportw
#undef inportb
#undef inportw
#endif



/*
this class controls the 8259A Programmable Interrupt Controllers
(PICs) on the Industry Standard Architecture (ISA) (IBM PC-AT clone)
motherboard

you only need to construct one of these (I suggest a global variable)
in your entire program, since it directly controls both PIC's

it automatically takes care of saving state on construction
and restoring it on destruction
*/


// manifest constants
//
#define NONSPECIFIC_END_OF_INTERRUPT    0x20
#define SPECIFIC_END_OF_INTERRUPT       0x60
#define SLAVE_8259A_PORT    0xA0
#define MASTER_8259A_PORT   0x20
#define INTR_MASK_OFFSET    1



// macros
//
#define irq_to_intr(level) ((level)>=8 ? 0x70+(level)-8 : 0x08+(level))



class InterruptController
{

private:
    // these save initial state
    //
    unsigned char old_master_mask;
    unsigned char old_slave_mask;

    // these shadow current state of chips
    //
    unsigned char master_mask;
    unsigned char slave_mask;

public:
    // constructor
    //
    InterruptController()
    {
        old_master_mask = inportb( MASTER_8259A_PORT + INTR_MASK_OFFSET );
        old_slave_mask = inportb( SLAVE_8259A_PORT + INTR_MASK_OFFSET );
        master_mask = old_master_mask;
        slave_mask = old_slave_mask;
    }

    // destructor
    //
    ~InterruptController()
    {
        outportb( MASTER_8259A_PORT + INTR_MASK_OFFSET, old_master_mask );
        outportb( SLAVE_8259A_PORT + INTR_MASK_OFFSET, old_slave_mask );
    }

    // disables a given IRQ level
    //
    void mask( unsigned level )
    {
        if( level >= 8 )
        {
            slave_mask |= 1 << (level - 8);
            outportb( SLAVE_8259A_PORT + INTR_MASK_OFFSET, slave_mask );

            // cascade pin IR2 on master PIC should not be disabled
            // because other IRQ levels on slave may need it
            //
        }
        else
        {
            master_mask |= 1 << level;
            outportb( MASTER_8259A_PORT + INTR_MASK_OFFSET, master_mask );
        }
    }

    // enables a given IRQ level
    //
    void unmask( unsigned level )
    {
        if( level >= 8 )
        {
            slave_mask &= ~(1 << (level - 8));
            outportb( SLAVE_8259A_PORT + INTR_MASK_OFFSET, slave_mask );

            // cascade pin IR2 on master PIC also needs to be
            // enabled, so just drop through to code for master
            //
            level = 2;
        }

        master_mask &= ~(1 << level);
        outportb( MASTER_8259A_PORT + INTR_MASK_OFFSET, master_mask );
    }

    // sends a nonspecific end of interrupt to the controller(s)
    // for the desired IRQ level
    //
    void static nonspecific_acknowledge( unsigned level )
    {
        if( level >= 8 )
            outportb( SLAVE_8259A_PORT, NONSPECIFIC_END_OF_INTERRUPT );

        outportb( MASTER_8259A_PORT, NONSPECIFIC_END_OF_INTERRUPT );
    }

    // sends a specific end of interrupt to the controller(s)
    // for the desired IRQ level
    //
    void static acknowledge( unsigned level )
    {
        disable();

        if( level >= 8 )
        {
            outportb( SLAVE_8259A_PORT, SPECIFIC_END_OF_INTERRUPT+level-8 );
            level = 2;
        }

        outportb( MASTER_8259A_PORT, SPECIFIC_END_OF_INTERRUPT+level );
    }
};


#endif      // not included yet
