#ifndef _MC146818_H
#define _MC146818_H



// headers we depend upon
//
#include <dos.h>                                // enable() disable() inportb() outportb()
#include "../types/types.h"                     // Bit8

#ifdef __DPMI32__
// linker won't resolve these unless we get rid of the
// header file's attempt to use use inline functions,
// which the 32-bit compiler doesn't support
//
#undef outportb
#undef outportw
#undef inportb
#undef inportw
#endif



// manifest constants
//
#define RTC_MAX_ADDRESS         127
#define RTC_USER_ADDRESS        10
#define RTC_USER_ADDRESS_COUNT  (RTC_MAX_ADDRESS + 1 - RTC_USER_ADDRESS)
#define RTC_ADDRESS_PORT        0x70
#define RTC_DATA_PORT           0x71



// CMOS real time clock chip, Motorola part# MC146818
//
// actually, this code only controls the hour, minute,
// and second time registers of the chip
//
// the hardware also has registers (which we ignore)
// for date of year, status, and general-purpose CMOS
// RAM for storage of system setup on AT-class machines
//
// only one instance of this class is allowed per
// computer, since there is only one of the hardware
// it models
//
// how do I guarantee that???
//
class RealTimeClock
{

public:
    // read one binary byte from the given address in CMOS RAM
    //
    static Bit8 get_one_binary( Bit8 address )
    {
        Bit8 masked;
        Bit8 value;

        // check the address range
        //
        // how could we return this error???
        //
        if( address > RTC_MAX_ADDRESS )
        {
            printf( "internal error!  RealTimeClock::get_one( address %d > %d )\n", address, RTC_MAX_ADDRESS );
            return 0;
        }

        // disable maskable interrupts
        //
        disable();

        // are nonmaskable interrupts currently masked?
        //
        masked = 0x80 & inportb( RTC_ADDRESS_PORT );

        // tell the chip the address we want
        //
        address &= RTC_MAX_ADDRESS;
        address |= masked;
        outportb( RTC_ADDRESS_PORT, address );

        // suck out the one byte we asked for
        //
        value = inportb( RTC_DATA_PORT );

        // enable maskable interrupts again
        //
        enable();

        return value;
    }

    // write the binary byte to the given address in CMOS RAM
    //
    static void put_one_binary( Bit8 address, Bit8 value )
    {
        Bit8 masked;

        // check the address range
        //
        // how could we return this error???
        //
        if( address > RTC_MAX_ADDRESS )
        {
            printf( "internal error!  RealTimeClock::put_one( address %d > %d )\n", address, RTC_MAX_ADDRESS );
            return;
        }

        // disable maskable interrupts
        //
        disable();

        // are nonmaskable interrupts currently masked?
        //
        masked = 0x80 & inportb( RTC_ADDRESS_PORT );

        // tell the chip the address we want
        //
        address &= RTC_MAX_ADDRESS;
        address |= masked;
        outportb( RTC_ADDRESS_PORT, address );

        // blow out the one byte we asked for
        //
        outportb( RTC_DATA_PORT, value );

        // enable maskable interrupts again
        //
        enable();
    }

    static void get_all_binary( Bit8 *array )
    {
        Bit8 address;

        for( address=0; address<RTC_USER_ADDRESS_COUNT; address++ )
            array[ address ] = get_one_binary( address + RTC_USER_ADDRESS );
    }

    static void put_all_binary( Bit8 *array )
    {
        Bit8 address;

        for( address=0; address<RTC_USER_ADDRESS_COUNT; address++ )
            put_one_binary( address+RTC_USER_ADDRESS, array[address] );
    }

    static Bit8 to_bcd( Bit8 value )
    {
        if( value >= 100 )
        {
            printf( "internal error: to_bcd( value %d > 100 )\n", value );
            return 0;
        }

        return ((value / 10) << 4) | (value % 10);
    }

    static Bit8 from_bcd( Bit8 value )
    {
        return (10 * ((value >> 4) & 0xF))
              + (value & 0xF);
    }

    // read one BCD byte from the given address
    // in CMOS RAM and convert it to binary
    //
    static Bit8 get_one_bcd( Bit8 address )
    {
        return from_bcd( get_one_binary( address ) );
    }

    // convert one byte from binary to BCD,
    // and write it to the given address in CMOS RAM
    //
    static void put_one_bcd( Bit8 address, Bit8 value )
    {
        put_one_binary( address, to_bcd( value ) );
    }

    // no constructor or destructor needed,
    // since there are no data elements
    //
    // having one of these would only serve as a shortcut,
    // since all the member functions are static, and hence
    // accessible without an instance of this class
    //

    // get the current time in seconds
    //
    // includes the minutes and hours also
    //
    static unsigned long get( void )
    {
        /* joel debug */
        //printf( "hour=%d\n", get_one( 4 ) );
        //printf( "minute=%d\n", get_one( 2 ) );
        //printf( "sec=%d\n", get_one( 0 ) );

        return (3600 * (unsigned long)get_one_bcd( 4 ))
               + (60 * (unsigned long)get_one_bcd( 2 ))
               +       (unsigned long)get_one_bcd( 0 );
    }
};



#endif      // not included yet
