#include <stdio.h>                  /* fprintf() stderr */
#include <stdlib.h>                 /* getenv() */
#include <math.h>                   /* atof() */
#include "../types/types.h"
#ifdef __unix__
#include <sys/time.h>               /* struct timeval */
#ifdef SUNOS
extern "C" int gettimeofday( struct timeval *tp, struct timezone *tzp );
#endif
#endif
#include "../intel/timestmp.h"


// conditional compilation switches
//
#define SHOW_CLOCK_RATE     TRUE


// manifest constants
//
#define CLOCK_RATE_ENVAR    "HOST_CLOCK_RATE"
#define NORMAL_CLOCK_BASE   1E6
#define NORMAL_CLOCK_MAX    1E9


// allocate storage for class-static variable
//
double PentiumClock::clock_rate = PentiumClock::init();


#ifdef CLK_86  // Use 80x86 clock hardware (rather than Pentium)

// CLK_86 code by Nevil Brownlee, Dec 96
//      Nevil's patches to fix midnight problems, Jun 97
//      Modified to use all 16 bits of timer0 counter, Dec 97
//      Change init() to call hwt_init() if neccessary, Jun 98

#include "..\dpmi32\phys.h"
#include <stdio.h>
#pragma inline
Bit32 *bios_tick_counter = 0;
#endif

extern "C" void hwt_init(void);
extern "C" void hwt_fin(void);


double PentiumClock::init()
{  // Called once only, to initialise class-static variables
#ifdef __unix__
    return 1e6;
#else
#ifdef CLK_86  // Use 80x86 clock hardware (rather than Pentium)
    hwt_init();
    return 1.19318E6;  /* PC clock rate */
#else
    char *clock_rate_string;

    /* get a pointer to the string for the environment variable
     * that holds the host CPU clock rate
     */
    clock_rate_string = getenv( CLOCK_RATE_ENVAR );
    if( clock_rate_string == NULL )
    {
        fprintf( stderr, "PentiumClock::init() error: environment variable '%s' is not set\n",
            CLOCK_RATE_ENVAR );
        throw 1;
    }

    /* convert the string to a double-precision floating-point number
     */
    clock_rate = atof( clock_rate_string );
    if( clock_rate == 0.0 )
    {
        fprintf( stderr, "PentiumClock::init() error: cannot convert environment variable '%s'"
            " contents '%s' to integer\n", CLOCK_RATE_ENVAR, clock_rate_string );
        throw 2;
    }

    /* check that the number is in a reasonable range for a CPU clock rate
     */
    if( clock_rate<NORMAL_CLOCK_BASE || clock_rate>NORMAL_CLOCK_MAX )
    {
        fprintf( stderr, "PentiumClock::init() error: environment variable '%s' contents"
            " %g are not in the range %g through %g\n", CLOCK_RATE_ENVAR,
            clock_rate, NORMAL_CLOCK_BASE, NORMAL_CLOCK_MAX );
        throw 3;
    }

#if SHOW_CLOCK_RATE
    printf( "PentiumClock::init() remembering host CPU clock rate is %.3f MHz\n", clock_rate/1E6 );
#endif

    return clock_rate;
#endif  /* defined( CLK_86 ) */
#endif  /* !defined( __unix__ ) */
}


// cannot be inline because it loops
//
Bit32 PentiumClock::wait( Bit32 count )
{
    PentiumClock until, now;

    now.set();
    until.low = count;
    until.high = 0;
    until += now;

    count = 0;

    do
    {
        now.set();
        count++;

    } while( (now.high==until.high && now.low<until.low)
          || (now.high<until.high) );

    return count;
}


#ifdef CLK_86  // Use 80x86 clock hardware (rather than Pentium)
static Bit32 last_tod;  /* Last observed BIOS time */
static Bit16 hwt_last_count;  /* Last observed timer0 value */

static Bit64_class day_ticks;  /* Whole days less init tod */

#define SEC_PER_DAY     86400L  // Seconds in a day
#define TICK_PER_DAY  1573040L  // BIOS ticks in a day
                 /* = 0x1800B0  */
#endif


// cannot be inline because there is no way to indicate
// to the compiler that it alters registers EAX and EDX
//
void PentiumClock::set( void )
{
#ifdef __unix__
    timeval now;
    gettimeofday( &now, NULL );
    low = now.tv_sec * 1000000;
    high = now.tv_sec - low / 1000000;
    low += now.tv_usec;
#else
#ifndef CLK_86  // Use the Pentium clock register
//    _EAX = 123;

    // read the Pentium 64-bit time stamp (processor clock) counter
    // into registers EDX (high 32 bits) and EAX (low 32 bits)
    //
    // must hand-assemble the RDTSC instruction, since MASM 5.10
    // doesn't know about anything past the 80386
    //
    // also, since the 32-bit compiler doesn't have a built-in
    // assembler at all, any embedded assembly causes it to restart
    // in run-my-output-through-TASM mode, which slows you down
    //
    // also, MASM generates 16-bit object files which TLINK32
    // doesn't like so you MUST either have TASM or hand-assemble
    // like this if you use BCC32
    //
    __emit__( 0x0F, 0x31 );

    // now copy the registers into the structure
    //
    low = _EAX;
    high = _EDX;

#else  // CLK_86 -> use the timer0 (BIOS) counter

   Bit16 count;
   Bit32 tod;
   asm pushf;  /* Save flags (especially int flag) */
   asm cli;    /* Disable interrupts */
      asm  mov   al,00H  /* Counter 0, Counter Latch */
      asm  out   43H,al
      asm  in    al,40H  /* Read LSB */
      asm  mov   ah,al
      asm  in    al,40H  /* Read MSB */
      asm  xchg  ah,al
      asm  dec   ax
      count = _AX;
      /* Interrupt occurs when counter changes from 1 to its initial
         value, i.e. 0.  We want count to be 0 then, and to increase
         to 0xFFFF just before the next interrupt.  */
      count = 0xFFFF - count;
      tod = *bios_tick_counter;

      if (tod == last_tod) {
         if (count < hwt_last_count)  /* One timer interrupt pending */
            last_tod = tod+1;
         }
      else if (tod > last_tod)
         last_tod = tod;
      else {  /* tod < last_tod */
         if (last_tod-tod > 0x180000) {   /* Midnight crossing */
            last_tod = tod;
            day_ticks += Bit64_class( (Bit32)(TICK_PER_DAY >> 16),
                                (Bit32)(TICK_PER_DAY << 16) );
            }
         if (count < hwt_last_count)  /* Several timer ints pending */
            ++last_tod;
         }
      hwt_last_count = count;

      this->low = last_tod << 16 | count;
      this->high = last_tod >> 16;
      *this += day_ticks;

   asm popf;   /* Restore flags */
#endif  /* defined( CLK_86 ) */
#endif  /* !defined( __unix__ ) */
   }


#ifndef CLK_86  // Initialise clock routines

extern "C" void hwt_init(void)
{ }
extern "C" void hwt_fin(void)
{ }

#else

static Bit32 real_mem_handle;  /* For accessing bottom 1M of real memory */

extern "C"
void hwt_init(void)  /* Initialise timer routines */
{
   Bit8 *real_mem, *seg0x40;
   if (bios_tick_counter) return;  /* Already initialised */
   // Only need access to the first page
#ifdef __DPMI32__
   real_mem = (Bit8 *)map_physical(0, 4096, &real_mem_handle);
   if (!real_mem) {
      printf("error mapping real mode memory\n");
      return;
      }
#else
   real_mem = (char *)0L;
#endif
   seg0x40 = &real_mem[0x400];  /* Access BIOS variables in real memory */
   bios_tick_counter = (Bit32 *)&seg0x40[0x6C];

   asm  mov  al,34H  /* Counter 0, LSB then MSB, mode 2 */
   asm  out  43H,al  /*    Decr in ones, interrupt every cycle */
   asm  mov  al,0
   asm  out  40H,al  /* Set initial count: */
   asm  out  40H,al  /*    LSB and MSB */

   hwt_last_count = 0;
   last_tod = *bios_tick_counter;
   day_ticks = Bit64_class( (Bit32)(~(last_tod >> 16)),
                      (Bit32)(~(last_tod << 16)) );
   day_ticks += Bit32(1);   /* day_ticks = -(save_last_tod << 16) */
   }

extern "C"
void hwt_fin(void)   /* Restore BIOS timer routines */
{
   asm  mov  al,36H  /* Counter 0, LSB+MSB, mode 3 */
   asm  out  43H,al  /*    Decr in twos, interrupt every second cycle */
   asm  mov  al,0
   asm  out  40H,al  /* Set initial count: */
   asm  out  40H,al  /*    LSB and MSB */

#ifdef __DPMI32__
   unmap_physical(4096, real_mem_handle);
#endif
   bios_tick_counter = NULL;
   }
#endif

/* The PC timer0 counter is driven at 1.19318 MHz, period 0.8380965 us

   256*period = 214.5527 us  <<< This was hw_time()'s resolution
                                 (earlier version using only MSB)

   65536*period = 54.925491 ms, BIOS clock ticks at 18.206482 Hz

   10 ms = 46.60859546 t0 ticks (10ms SNMP ticks)
   (512/11 = 46.545  0.01% conversion error)  */

