#include <stdio.h>                      // printf()
#include <stdlib.h>                     // exit()
#include <io.h>                         // open() ioctl()
#include <fcntl.h>                      // O_RDONLY O_BINARY
#include <errno.h>                      // errno
#include <stdarg.h>                     // va_list va_start() va_end()
#include <mem.h>                        // memset()
#include "../intel/timestmp.h"          // PentiumClock
#include "../scsi/aspi.h"               // prototypes we have to adhere to


static AspiHandler aspi_init( void )
{
    int handle;
    int result;
    AspiHandler aspi0;

    // open the ASPI device driver, which DOS only knows about
    // so it can give us the entrypoint via an IOCTL
    //
    handle = open( "SCSIMGR$", O_RDONLY|O_BINARY );
    if( handle < 0 )
    {
        fprintf( stderr, "error %d '%s' opening SCSIMGR$\naborting\n",
            errno, strerror(errno) );
        exit( 1 );
    }

    // ask the ASPI manager for its directly-callable entrypoint
    //
    result = ioctl( handle, 2, &aspi0, sizeof(aspi0) );
    if( result < 0 )
    {
        fprintf( stderr, "error %d '%s' reading %d byte entrypoint from ASPI\naborting\n",
            sizeof(aspi0), errno, strerror(errno) );
        exit( 2 );
    }
    if( result != sizeof( aspi0 ) )
    {
        fprintf( stderr, "fatal error: only read %d of %d byte entrypoint from ASPI\naborting\n",
            result, sizeof(aspi0), errno, strerror(errno) );
        exit( 3 );
    }

    close( handle );

    return aspi0;
}



AspiHandler aspi = aspi_init();



// returns 0  if successful
// returns 1  if error we were told to ignore
// returns -1 if timeout
// returns -2 if any other error
//
extern "C"
int aspi_request_and_wait(
    AspiSrb *my_srb,
    AspiStatus ignore_me,
    double timeout,
    double *elapsed,
    int always_show_timing,
    char *format,
    ...
)
{
    va_list argptr;
    PentiumClock start, alarm, now;
    int result;

    // remember when we started this request
    //
    alarm.put_seconds( timeout );
    start.set();
    alarm += start;

    aspi( my_srb );

    // wait for the driver to complete the command
    //
    do
    {
        now.set();

    } while( my_srb->GetDeviceType.Status==AspiStatusScsiRequestInProgress
    && now < alarm );

    va_start( argptr, format );

    // deal with errors and invalid adaptors
    //
    if( my_srb->GetDeviceType.Status == AspiStatusScsiRequestInProgress )
    {
        printf( "timed out " );
        vprintf( format, argptr );
        my_srb->GetDeviceType.Status = AspiStatusTimeout;
        result = -1;
    }
    else
    {
        now -= start;
        if( always_show_timing )
            printf( "\nwaited %g seconds for SCSI request to complete\n", now.get_seconds() );
        if( elapsed )
            *elapsed = now.get_seconds();

        if( my_srb->GetDeviceType.Status != AspiStatusScsiRequestCompletedWithoutError )
        {
            // Adaptec ASPI8DOS.SYS version 1.32 doesn't return an error
            // in the status field when you give it an invalid adaptor
            // number; instead we must time out or avoid the situation...
            //
            if( my_srb->GetDeviceType.Status != ignore_me )
            {
                printf( "status %d ", my_srb->GetDeviceType.Status );
                vprintf( format, argptr );
                result = 1;
            }
            else
                result = -2;
        }
        else
            result = 0;
    }

    va_end( argptr );
    return result;
}




#define FIND_DISK_TIMEOUT         1.0
#define MAX_DISKS                 100
extern unsigned aspi_disk_count = 0;
AspiDisk aspi_all_disks[ MAX_DISKS ];



int aspi_find_disks( AspiSrb *my_srb, unsigned max_per_bus, Bit16 max_buses )
{
    AspiHostAdaptorInquiry *my_inquiry = &my_srb->HostAdaptorInquiry;
    AspiGetDeviceType *my_getdev = &my_srb->GetDeviceType;
    AspiGetDiskDriveInformation *my_getdisk = &my_srb->GetDiskDriveInformation;
    Bit16 adaptor_count;
    unsigned parm_count;
    Bit16 scsi_id_count;
    Bit16 scsi_lun_count;
    unsigned disks_on_this_bus;

    aspi_disk_count = 0;

    for( adaptor_count=0; adaptor_count<max_buses; adaptor_count++ )
    {
        // default all fields to zero
        //
        memset( my_srb, 0, sizeof(AspiSrb) );

        // say we want info on this host adaptor
        //
        my_inquiry->Command = AspiCommandHostAdaptorInquiry;
        my_inquiry->HostAdaptorNumber = adaptor_count;
        my_inquiry->ExtendedRequestSignature = AspiExtendedRequestSignature;
        my_inquiry->ExtendedBufferLength = 4;

        if( aspi_request_and_wait( my_srb,
            AspiStatusInvalidHostAdaptorNumber,
            FIND_DISK_TIMEOUT, NULL, TRUE,
            "getting info on host adaptor %d", adaptor_count
        ) )
            continue;

        // decode the returned data
        //
        printf( "host adaptor %d said...\n", adaptor_count );

        printf( "%d host adaptors\n", my_inquiry->NumberOfHostAdaptors );
        if( max_buses > my_inquiry->NumberOfHostAdaptors )
            max_buses = my_inquiry->NumberOfHostAdaptors;

        printf( "host adaptor SCSI ID is %d\n", my_inquiry->HostAdaptorScsiId );
        printf( "SCSI manager description is '%.16s'\n", my_inquiry->ScsiManagerDescription );
        printf( "host adaptor description is '%.16s'\n", my_inquiry->HostAdaptorDescription );
        printf( "host adaptor unique parameters are '" );
        for( parm_count=0; parm_count<sizeof(my_inquiry->HostAdaptorUniqueParameters); parm_count++ )
            printf( "%.2X ", my_inquiry->HostAdaptorUniqueParameters[parm_count] );
        printf( "'\n" );

        if( my_inquiry->ExtendedRequestSignature != AspiExtendedAcceptedSignature )
            printf( "ASPI did not accept the extended request\n" );
        else
        {
            printf( "ASPI returned %d extended bytes\n", my_inquiry->ExtendedBufferLength );
            printf( "residual byte length can%s be reported\n",
                my_inquiry->SupportedExtensions.ResidualByteLengthReported ? "" : " not" );
            printf( "host adaptor is%s wide SCSI 16 bit\n",
                my_inquiry->SupportedExtensions.WideScsi16BitHostAdaptor ? "" : " not" );
            printf( "host adaptor is%s wide SCSI 32 bit\n",
                my_inquiry->SupportedExtensions.WideScsi32BitHostAdaptor ? "" : " not" );
        }

        disks_on_this_bus = 0;

        // for all possible devices on a SCSI bus...
        //
        for( scsi_id_count=0; scsi_id_count<=MAX_SCSI_ID; scsi_id_count++ )
        {
            for( scsi_lun_count=0; scsi_lun_count<=MAX_SCSI_LUN; scsi_lun_count++ )
            {
                // default all fields to zero
                //
                memset( my_srb, 0, sizeof(AspiSrb) );

                // say we want info on this host adaptor
                //
                my_getdev->Command = AspiCommandGetDeviceType;
                my_getdev->HostAdaptorNumber = adaptor_count;
                my_getdev->TargetId = scsi_id_count;
                my_getdev->Lun = scsi_lun_count;

                if( aspi_request_and_wait( my_srb,
                    AspiStatusScsiDeviceNotInstalled,
                    FIND_DISK_TIMEOUT, NULL, FALSE,
                    "getting device type for SCSI bus %d ID %d and LUN %d\n",
                    adaptor_count, scsi_id_count, scsi_lun_count
                ) )
                    continue;

                // decode the returned data
                //
                if( my_getdev->PeripheralDeviceType != AspiPeripheralTypeHardDisk )
                {
                    printf( "SCSI bus %d ID %d LUN %d is peripheral type %d\n",
                        adaptor_count, scsi_id_count, scsi_lun_count,
                        my_getdev->PeripheralDeviceType
                    );
                }
                else
                {
                    // get info for the harddrive
                    //
                    memset( my_srb, 0, sizeof(AspiSrb) );
                    my_getdisk->Command = AspiCommandGetDiskDriveInformation;
                    my_getdisk->HostAdaptorNumber = adaptor_count;
                    my_getdisk->TargetId = scsi_id_count;
                    my_getdisk->Lun = scsi_lun_count;

                    if( !aspi_request_and_wait( my_srb,
                        AspiStatusIgnoreNothing,
                        FIND_DISK_TIMEOUT, NULL, FALSE,
                        "getting disk info for SCSI bus %d ID %d and LUN %d\n",
                        adaptor_count, scsi_id_count, scsi_lun_count
                    ) )
                    {
                        // decode the returned data
                        //
                        printf( "\nSCSI bus %d ID %d LUN %d disk info...\n",
                            adaptor_count, scsi_id_count, scsi_lun_count );

                        switch( my_getdisk->DriveFlags.Int13Info )
                        {
                        case AspiInt13Invalid:
                            printf( "invalid INT 13h drive accessibility returned!\n" );
                            break;

                        case AspiInt13Inaccessible:
                            printf( "not accessible using BIOS INT 13h\n" );
                            break;

                        case AspiInt13BiosAndDos:
                            printf( "drive is controlled by DOS and accessible"
                                " through BIOS INT 13h as 0x%.2X\n", my_getdisk->Int13Drive );
                            break;

                        case AspiInt13BiosOnly:
                            printf( "drive is not controlled by DOS but is accessible"
                                " through BIOS INT 13h as 0x%.2X\n", my_getdisk->Int13Drive );
                            break;
                        }

                        printf( "preferred head translation is %d\n", my_getdisk->PreferredHeadTranslation );
                        printf( "preferred sector translation is %d\n", my_getdisk->PreferredSectorTranslation );

                        aspi_all_disks[ aspi_disk_count ].HostAdaptorNumber = adaptor_count;
                        aspi_all_disks[ aspi_disk_count ].TargetId          = scsi_id_count;
                        aspi_all_disks[ aspi_disk_count ].Lun               = scsi_lun_count;
                        aspi_disk_count++;

                        if( ++disks_on_this_bus == max_per_bus )
                            goto bus_done;      // like 2 break's and a continue
                    }
                }
            }
        }

        bus_done:
    }

    return 0;
}



