#ifndef _ASPI_H
#define _ASPI_H



// ASPI = Advanced SCSI Programming Interface
// SCSI = Small Computer Systems Interface
// LUN  = Logical Unit Number (within a SCSI target ID)
// SRB  = SCSI Request Block
// CDB  = Command Descriptor Block (sent to SCSI device, see SCSI.H for details)
//

#include "../scsi/scsi.h"

#ifdef __cplusplus
extern "C" {
#endif

#if __DPMI32__
#   define far
#   define _cdecl
#endif



typedef enum
{
    AspiCommandHostAdaptorInquiry = 0,
    AspiCommandGetDeviceType = 1,
    AspiCommandExecuteScsiIo = 2,
    AspiCommandAbortScsiIo = 3,
    AspiCommandResetScsiDevice = 4,
    AspiCommandSetHostAdaptorParameters = 5,
    AspiCommandGetDiskDriveInformation = 6,
    AspiCommandReservedForFutureExpansionFirst = 7,
    AspiCommandReservedForFutureExpansionLast = 127,
    AspiCommandReservedForVendorUniqueFirst = 128,
    AspiCommandReservedForVendorUniqueLast = 255
} AspiCommand;

typedef enum
{
    AspiStatusScsiRequestInProgress = 0,
    AspiStatusScsiRequestCompletedWithoutError = 1,
    AspiStatusScsiRequestAbortedByHost = 2,
    AspiStatusScsiRequestCompletedWithError = 4,
    AspiStatusInvalidScsiRequest = 128,
    AspiStatusInvalidHostAdaptorNumber = 129,
    AspiStatusScsiDeviceNotInstalled = 130,
    AspiStatusTimeout = 255,                // artificially inserted by my code
    AspiStatusIgnoreNothing = 256           // artificially passed to my code
} AspiStatus;

typedef enum
{
    AspiAdaptorStatusNoneDetected = 0,
    AspiAdaptorStatusSelectionTimeout = 17,
    AspiAdaptorStatusDataOverrunOrUnderrun = 18,
    AspiAdaptorStatusUnexpectedBusFree = 19,
    AspiAdaptorStatusTargetBusPhaseSequenceFailure = 20
} AspiAdaptorStatus;

typedef enum
{
    AspiTargetStatusNoneAvailable = 0,
    AspiTargetStatusCheck = 2,                  // see sense data for details
    AspiTargetStatusTargetOrLunBusy = 8,        // see sense data for details
    AspiTargetStatusReservationConflict = 24
} AspiTargetStatus;

typedef enum
{
    AspiXferDirectionFromScsiCommand  = 0,      // length not checked
    AspiXferDirectionScsiTargetToHost = 1,      // length checked
    AspiXferDirectionHostToScsiTarget = 2,      // length checked
    AspiXferDirectionNoDataTransfer   = 3       // who cares about length anyways?
} AspiXferDirection;

// only applicable to AspiCommandExecuteScsiIo
// and AspiCommandResetScsiDevice, and the latter
// only uses the EnablePosting field
//
typedef struct
{
    Bit8 EnablePosting      : 1;
    Bit8 EnableLinking      : 1;                // must be FALSE if scatter/gather used
    Bit8 EnableResidualByteLengthReporting : 1; // also requires driver support
    Bit8 TransferDirection  : 2;                // see AspiXferDirection above
    Bit8 Reserved           : 3;
} AspiScsiRequestFlags;



// W = sent to ASPI
// R = returned from ASPI
// - = reserved and should be sent to ASPI as 0
// ? = not documented who fills in and why
// J = used only by Joel's functions
//

typedef enum
{
    AspiExtendedRequestSignature = 0xAA55,
    AspiExtendedAcceptedSignature = 0x55AA
} AspiExtended;

// executes synchronously, so you can assume result is
// available immediately after aspi() returns to you
//
typedef struct
{
    Bit8 Command;                               // W    0 for this command
    Bit8 Status;                                // R
    Bit8 HostAdaptorNumber;                     // W    0-based
    Bit8 ScsiRequestFlags;                      // -
    Bit16 ExtendedRequestSignature;             // R/W  client sets to 0xAA55, driver sets to 0x55AA if extended request is supported
    Bit16 ExtendedBufferLength;                 // R/W  client sets to 4, driver returns (possibly smaller) number of valid bytes
    Bit8 NumberOfHostAdaptors;                  // R    1-based
    Bit8 HostAdaptorScsiId;                     // R    0-based
    Bit8 ScsiManagerDescription[ 16 ];          // R    ASCII string
    Bit8 HostAdaptorDescription[ 16 ];          // R    ASCII string
    Bit8 HostAdaptorUniqueParameters[ 16 ];     // ?    implementation-specific
    struct
    {
        Bit16 Reserved1                     : 1;
        Bit16 ResidualByteLengthReported    : 1;
        Bit16 WideScsi16BitHostAdaptor      : 1;
        Bit16 WideScsi32BitHostAdaptor      : 1;
        Bit16 Reserved2                     : 12;
    } SupportedExtensions;                      // R
} AspiHostAdaptorInquiry;

// these are supposed to be documented along with the
// SCSI mode inquiry command, but I don't have a SCSI
// manual from Seagate yet, so this is the only one
// I have seen so far
//
typedef enum
{
    AspiPeripheralTypeHardDisk = 0
} AspiPeripheralType;

// executes synchronously, so you can assume result is
// available immediately after aspi() returns to you
//
typedef struct
{
    Bit8 Command;                               // W    1 for this command
    Bit8 Status;                                // R    if AspiStatusScsiDeviceNotInstalled, PeripheralDeviceType was not filled in by driver
    Bit8 HostAdaptorNumber;                     // W    0-based
    Bit8 ScsiRequestFlags;                      // -
    Bit32 ReservedForExpansion;                 // -
    Bit8 TargetId;                              // W
    Bit8 Lun;                                   // W
    Bit8 PeripheralDeviceType;                  // R    previously reported by SCSI inquiry command
} AspiGetDeviceType;

// executes asynchronously, so either poll the status
// field or device driver can call you back when done
//
typedef struct
{
    Bit8 Command;                               // W    2 for this command
    Bit8 Status;                                // R
    Bit8 HostAdaptorNumber;                     // W    0-based
    AspiScsiRequestFlags RequestFlags;          // W
    Bit32 ReservedForExpansion;                 // -
    Bit8 TargetId;                              // W
    Bit8 Lun;                                   // W
    Bit32 DataAllocationLength;                 // R/W  little-endian, must be zero for SCSI commands that don't transfer data, like "rewind" and "start unit"
    Bit8 SenseAllocationLength;                 // W    how many bytes allocated at end of SRB for sense data (255 already allocated in this structure)
    Bit32 DataBufferPointer;                    // W    realmode seg:ofs logical, not physical
    Bit32 SrbLinkPointer;                       // W    realmode seg:ofs logical, not physical
                                                //      (used to guarantee seq execution of SCSI commands ??on device??)
                                                //      requires AspiScsiRequestFlags.EnableLinking==TRUE and linking requested by SCSI CDB and linking supported by SCSI target/LUN
                                                //      !!!strongly discouraged by Adaptec since many SCSI devices don't support linking
    Bit8 CdbLength;                             // W    filled-in bytes, 0-based (I guess)
    Bit8 HostAdaptorStatus;                     // R    see enum AspiAdaptorStatus
    Bit8 TargetStatus;                          // R    see enum AspiTargetStatus
    Bit32 PostRoutineAddress;                   // W    realmode seg:ofs logical, not physical
    Bit8 ReservedForAspiWorkspace[ 34 ];        // -
    union
    {
        ScsiCdb Cdb;                            // W    client puts CDB here
        ScsiSense Sense;                        // R    driver puts sense data here (by doing a request sense SCSI command) if check condition is present at end of SCSI command
        Bit8 Padding[ 256 ];                    //      probably not necessary, but makes sure max length is allocated
    } CdbAndSense;                              // R/W
} AspiExecuteScsiIo;

// returns success in this SRB immediately, but
// also causes referred-to SRB to complete later
//
// cannot be issued during a post routine!
//
typedef struct
{
    Bit8 Command;                               // W    3 for this command
    Bit8 Status;                                // R
    Bit8 HostAdaptorNumber;                     // W    0-based
    Bit8 ScsiRequestFlags;                      // -
    Bit32 ReservedForExpansion;                 // -
    Bit32 SrbPointer;                           // W    realmode seg:ofs logical, not physical
} AspiAbortScsiIo;

// executes asynchronously, so either poll the status
// field or device driver can call you back when done
//
typedef struct
{
    Bit8 Command;                               // W    4 for this command
    Bit8 Status;                                // R
    Bit8 HostAdaptorNumber;                     // W    0-based
    Bit8 ScsiRequestFlags;                      // W    see struct AspiScsiRequestFlags above
    Bit32 ReservedForExpansion;                 // -
    Bit8 TargetId;                              // W
    Bit8 Lun;                                   // W
    Bit8 Reserved2[ 14 ];                       // -
    Bit8 HostAdaptorStatus;                     // R    see enum AspiAdaptorStatus
    Bit8 TargetStatus;                          // R    see enum AspiTargetStatus
    Bit32 PostRoutineAddress;                   // W    realmode seg:ofs logical, not physical
    Bit8 ReservedForAspiWorkspace[ 2 ];         // -
} AspiResetScsiDevice;

// executes synchronously, so you can assume result is
// available immediately after aspi() returns to you
//
typedef struct
{
    Bit8 Command;                               // W    5 for this command
    Bit8 Status;                                // R
    Bit8 HostAdaptorNumber;                     // W    0-based
    Bit8 ScsiRequestFlags;                      // -
    Bit32 ReservedForExpansion;                 // -
    Bit8 HostAdaptorUniqueParameters[ 16 ];     // ?    implementation-specific
} AspiSetHostParameters;

typedef enum
{
    AspiInt13Inaccessible = 0,                  // Int13Drive field is invalid, cannot see from either BIOS or DOS
    AspiInt13BiosAndDos = 1,                    // Int13Drive field is filled in and drive is controlled by DOS
    AspiInt13BiosOnly = 2,                      // Int13Drive field is filled in but drive is not controlled by DOS
    AspiInt13Invalid = 3                        // not a valid value
} AspiInt13;

// executes synchronously, so you can assume result is
// available immediately after aspi() returns to you
//
typedef struct
{
    Bit8 Command;                               // W    6 for this command
    Bit8 Status;                                // R    always AspiStatusScsiRequestCompletedWithoutError if supported, otherwise AspiStatusInvalidScsiRequest
    Bit8 HostAdaptorNumber;                     // W    0-based
    Bit8 ScsiRequestFlags;                      // -
    Bit32 ReservedForExpansion;                 // -
    Bit8 TargetId;                              // W
    Bit8 Lun;                                   // W
    struct
    {
        Bit8 Int13Info : 2;
        Bit8 Reserved  : 6;                     //      see enum AspiInt13
    } DriveFlags;                               // R
    Bit8 Int13Drive;                            // R    what drive number BIOS sees this target/LUN as
    Bit8 PreferredHeadTranslation;              // R    host adaptor's / disk drive's preferred head count
    Bit8 PreferredSectorTranslation;            // R    host adaptor's / disk drive's preferred sectors per track
    Bit8 Reserved[ 10 ];                        // -
} AspiGetDiskDriveInformation;



typedef union
{
    AspiHostAdaptorInquiry HostAdaptorInquiry;
    AspiGetDeviceType GetDeviceType;
    AspiExecuteScsiIo ExecuteScsiIo;
    AspiAbortScsiIo AbortScsiIo;
    AspiResetScsiDevice ResetScsiDevice;
    AspiSetHostParameters SetHostParameters;
    AspiGetDiskDriveInformation GetDiskDriveInformation;
} AspiSrb;


// both client post handlers and the ASPI driver's
// entrypoint use this sort of function
//
// caller pops args
//
typedef void far _cdecl (*AspiHandler)(AspiSrb far *srb);

// ASPI driver's entrypoint, determined once during init
//
extern AspiHandler aspi;

// helper functuions I wrote myself,
// and their data structures
//

int aspi_request_and_wait(
    AspiSrb *my_srb,
    AspiStatus ignore_me,
    double timeout,
    double *elapsed,
    int always_show_timing,
    char *format,
    ...
);

typedef struct
{
    Bit8 HostAdaptorNumber;
    Bit8 TargetId;
    Bit8 Lun;
} AspiDisk;

extern unsigned aspi_disk_count;
extern AspiDisk aspi_all_disks[];
int aspi_find_disks( AspiSrb *my_srb, unsigned max_per_bus, Bit16 max_buses );



#ifdef __cplusplus
}
#endif



#endif      // not included yet
