#ifndef _SCSI_H
#define _SCSI_H



// SCSI = Small Computer Systems Interface
// LUN  = Logical Unit Number (within a SCSI target ID)
// CDB  = Command Descriptor Block (sent to SCSI device, see SCSI.H for details)
// RPL  = ???
// APM  = Advanced Power Management???
//

#include "../types/types.h"

#ifdef __cplusplus
extern "C" {
#endif



#define MAX_SCSI_ID                 15
#define MAX_SCSI_LUN                7
#define MAX_SCSI_DEVICES_PER_BUS    ((MAX_SCSI_ID + 1) * (MAX_SCSI_LUN + 1))
#define MAX_SCSI_BUSES              256             // 3 per card * 4 slots is more reasonable



// see from Seagate Disc Driver SCSI2/SCSI3 Interface manual
// Volume 2 Version 2 Rev H, part number 77738479, (c) 1997
// page 48, section 4.2.1 "Operation Code"
//
// upper 3 bits of SCSI operation codes are called "group code"
//    group    0: 6 byte command         0-31   00-1Fh
//    groups 1,2: 10 byte commands      32-95   20-5Fh
//    groups 3,4: reserved              96-159  60-9Fh
//    group    5: 12 byte commands     160-191  A0-BFh
//    groups 6,7: vendor specific      192-255  C0-FFh
//
typedef enum
{
    ScsiCommandTestUnitReady = 0x00,
    ScsiCommandRezeroUnit = 0x01,                   // seek to beg of disk, do thermal comp, save log
    ScsiCommandRequestSense = 0x03,
    ScsiCommandFormatUnit = 0x04,
    ScsiCommandReassignBlocks = 0x07,
    ScsiCommandRead = 0x08,                         // 21-bit LBA
    ScsiCommandWrite = 0x0A,                        // 21-bit LBA
    ScsiCommandSeek = 0x0B,                         // 21-bit LBA
    ScsiCommandInquiry = 0x12,
    ScsiCommandModeSelect6 = 0x15,
    ScsiCommandReserve = 0x16,
    ScsiCommandRelease = 0x17,
    ScsiCommandCopy = 0x18,                         // not implemented
    ScsiCommandModeSense6 = 0x1A,
    ScsiCommandStartOrStopUnit = 0x1B,
    ScsiCommandReceiveDiagnosticResults = 0x1C,
    ScsiCommandSendDiagnostic = 0x1D,
    ScsiCommandPreventOrAllowMediumRemoval = 0x1E,  // not implemented
    ScsiCommandReadCapacity = 0x25,
    ScsiCommandReadExtended = 0x28,                 // what I want to use
    ScsiCommandWriteExtended = 0x2A,                // what I want to use
    ScsiCommandSeekExtended = 0x2B,
    ScsiCommandWriteAndVerify = 0x2E,
    ScsiCommandVerify = 0x2F,
    ScsiCommandPrefetch = 0x34,
    ScsiCommandSynchronizeCache = 0x35,
    ScsiCommandReadDefectData = 0x37,
    ScsiCommandCompare = 0x39,                      // not implemented
    ScsiCommandCopyAndVerify = 0x3A,                // not implemented
    ScsiCommandWriteBuffer = 0x3B,                  // for host diags, also downloads microcode
    ScsiCommandReadBuffer = 0x3C,                   // for host diags
    ScsiCommandReadLong = 0x3E,                     // gets data and ECC bytes for a single block
    ScsiCommandWriteLong = 0x3F,                    // puts data and ECC bytes for a single block
    ScsiCommandChangeDefinition = 0x40,             // SCSI1 or SCSI2/3 mode
    ScsiCommandWriteSame = 0x41,                    // good for zeroing up to end of disk
    ScsiCommandLogSelect = 0x4C,
    ScsiCommandLogSense = 0x4D,
    ScsiCommandXdWrite = 0x50,
    ScsiCommandXpWrite = 0x51,
    ScsiCommandXdRead = 0x52,
    ScsiCommandModeSelect10 = 0x55,
    ScsiCommandReserve10 = 0x56,
    ScsiCommandRelease10 = 0x57,
    ScsiCommandModeSense10 = 0x5A,
    ScsiCommandXdWriteExtended = 0x80,
    ScsiCommandRebuild = 0x81,
    ScsiCommandRegenerate = 0x82,
} ScsiCommand;

typedef enum
{
    ScsiErrorCurrent = 0x70,
    ScsiErrorDeferred = 0x71                    // see page 86, section 5.1.1.2.1
} ScsiErrorCode;

// from page 83, table 5.1.1-5
//
typedef enum
{
    ScsiSenseKeyNoSense = 0,                    // successful command or ILI bit is 1
    ScsiSenseKeyRecoveredError = 1,             // for some mode settings, last cmd may have terminated before completing
    ScsiSenseKeyNotReady = 2,
    ScsiSenseKeyMediumError = 3,
    ScsiSenseKeyHardwareError = 4,              // nonrecoverable h/w error during cmd or selftest, like i/f parity err, controller failure, device failure, etc.
    ScsiSenseKeyIllegalRequest = 5,             // if in CDB, media unaltered, else if in addl. parms supplied as data, media may be altered already.
                                                // may indicate invalid IDENTIFY msg rcvd or attempt to write past last logical block
    ScsiSenseKeyUnitAttention = 6,              // drive may have been reset; see UnitAttention command in section 4.6
    ScsiSenseKeyDataProtect = 7,
    ScsiSenseKeyUnassigned1 = 8,
    ScsiSenseKeyFirmwareError = 9,              // vendor specific
    ScsiSenseKeyUnassigned2 = 10,
    ScsiSenseKeyAbortedCommand = 11,            // drive aborted the cmd, but initiator may be able to recover by retrying
    ScsiSenseKeyEqual = 12,                     // SEARCH DATA command satisfied an equal comparison
    ScsiSenseKeyVolumeOverflow = 13,            // buffered peripheral device reached end of medium partition and some data in buffer hasn't been written
    ScsiSenseKeyMiscompare = 14,                // source data didn't match data read from medium
    ScsiSenseKeyUnassigned3 = 15
} ScsiSenseKey;

// 13th (msb) and 14th (lsb) bytes of extended sense data
//
typedef enum
{
    ScsiAdditionalSenseCodeNoAdditionalSenseCodeInformation = 0x0000,
    ScsiAdditionalSenseCodeNoIndexOrLogicalBlockSignal = 0x0100,
    ScsiAdditionalSenseCodeNoSeekComplete = 0x0200,
    ScsiAdditionalSenseCodePeripheralDeviceWriteFault = 0x0300,
    ScsiAdditionalSenseCodeLogicalUnitNotReadyAndCauseNotReportable = 0x0400,
    ScsiAdditionalSenseCodeLogicalUnitNotReadyAndBecomingReady = 0x0401,
    ScsiAdditionalSenseCodeLogicalUnitNotReadyAndInitializingCommandRequired = 0x0402,
    ScsiAdditionalSenseCodeLogicalUnitNotReadyAndManualInterventionRequired = 0x0403,
    ScsiAdditionalSenseCodeLogicalUnitNotReadyAndFormatInProgress = 0x0404,
    ScsiAdditionalSenseCodeLogicalUnitCommunicationFailure = 0x0800,
    ScsiAdditionalSenseCodeLogicalUnitCommunicationTimeout = 0x0801,
    ScsiAdditionalSenseCodeLogicalUnitCommunicationParityError = 0x0802,
    ScsiAdditionalSenseCodeTrackFollowingError = 0x0900,
    ScsiAdditionalSenseCodeServoFault = 0x0901,
    ScsiAdditionalSenseCodeHeadSelectFault = 0x0904,
    ScsiAdditionalSenseCodeErrorLogOverflow = 0x0A00,
    ScsiAdditionalSenseCodeWriteError = 0x0C00,
    ScsiAdditionalSenseCodeWriteErrorRecoveredWithAutoReallocation = 0x0C01,
    ScsiAdditionalSenseCodeWriteErrorAutoReallocationFailed = 0x0C02,
    ScsiAdditionalSenseCodeIdCrcOrEccError = 0x1000,
    ScsiAdditionalSenseCodeUnrecoveredReadError = 0x1100,
    ScsiAdditionalSenseCodeReadRetriesExhausted = 0x1101,
    ScsiAdditionalSenseCodeErrorTooLongToCorrect = 0x1102,
    ScsiAdditionalSenseCodeUnrecoveredReadErrorAndAutoReallocationFailed = 0x1104,
    ScsiAdditionalSenseCodeAddressMarkNotFoundForIdField = 0x1200,
    ScsiAdditionalSenseCodeRecoveredDataWithoutEccUsingPreviousLogicalBlockId = 0x1201,
    ScsiAdditionalSenseCodeRecoveredDataWithEccUsingPreviousLogicalBlockId = 0x1202,
    ScsiAdditionalSenseCodeLogicalBlockNotFound = 0x1400,
    ScsiAdditionalSenseCodeRecordNotFound = 0x1401,
    ScsiAdditionalSenseCodeRandomPositioningError = 0x1500,
    ScsiAdditionalSenseCodeMechanicalPositioningError = 0x1501,
    ScsiAdditionalSenseCodePositioningErrorDetectedByReadOfMedium = 0x1502,
    ScsiAdditionalSenseCodeDataSynchronizationMarkError = 0x1600,
    ScsiAdditionalSenseCodeRecoveredDataWithNoErrorCorrectionApplied = 0x1700,
    ScsiAdditionalSenseCodeRecoveredDataUsingRetries = 0x1701,
    ScsiAdditionalSenseCodeRecoveredDataUsingPositiveOffset = 0x1702,
    ScsiAdditionalSenseCodeRecoveredDataUsingNegativeOffset = 0x1703,
    ScsiAdditionalSenseCodeRecoveredDataUsingPreviousLogicalBlockId = 0x1705,
    ScsiAdditionalSenseCodeRecoveredDataWithoutEccAndDataAutoReallocated = 0x1706,
    ScsiAdditionalSenseCodeRecoveredDataWithEcc = 0x1800,
    ScsiAdditionalSenseCodeRecoveredDataWithEccAndRetriesApplied = 0x1801,
    ScsiAdditionalSenseCodeRecoveredDataWithEccAndOrRetriesAndDataAutoReallocated = 0x1802,
    ScsiAdditionalSenseCodeRecoveredDataWithEccAndOrRetries = 0x1805,
    ScsiAdditionalSenseCodeDefectListError = 0x1900,
    ScsiAdditionalSenseCodeDefectListNotAvailable = 0x1901,
    ScsiAdditionalSenseCodeDefectListErrorInPrimaryList = 0x1902,
    ScsiAdditionalSenseCodeDefectListErrorInGrownList = 0x1903,
    ScsiAdditionalSenseCodeParameterListLengthError = 0x1A00,
    ScsiAdditionalSenseCodeSynchronousDataTransferError = 0x1B00,
    ScsiAdditionalSenseCodeDefectListNotFound = 0x1C00,
    ScsiAdditionalSenseCodePrimaryDefectListNotFound = 0x1C01,
    ScsiAdditionalSenseCodeGrownDefectListNotFound = 0x1C02,
    ScsiAdditionalSenseCodeSeagateUniqueDiagnosticCode = 0x1C83,
    ScsiAdditionalSenseCodeMiscompareDuringVerifyOperation = 0x1D00,
    ScsiAdditionalSenseCodeInvalidCommandOperationCode = 0x2000,
    ScsiAdditionalSenseCodeLogicalBlockAddressOutOfRange = 0x2100,
    ScsiAdditionalSenseCodeInvalidFieldInCdb = 0x2400,
    ScsiAdditionalSenseCodeLogicalUnitNotSupported = 0x2500,
    ScsiAdditionalSenseCodeInvalidFieldInParameterList = 0x2600,
    ScsiAdditionalSenseCodeParameterNotSupported = 0x2601,
    ScsiAdditionalSenseCodeParameterValueInvalid = 0x2602,
    ScsiAdditionalSenseCodeInvalidFieldParameterIsThreshold = 0x2603,
    ScsiAdditionalSenseCodeInvalidFieldParameterIsChecksum = 0x2698,
    ScsiAdditionalSenseCodeInvalidFieldParameterIsFirmwareTag = 0x2699,
    ScsiAdditionalSenseCodeWriteProtected = 0x2700,
    ScsiAdditionalSenseCodeNotReadyToReadyTransitionSoMediaMayHaveChanged = 0x2800,
    ScsiAdditionalSenseCodePowerOnOrResetOrBusDeviceResetOccurred = 0x2900,
    ScsiAdditionalSenseCodePowerOnResetOccurred = 0x2901,
    ScsiAdditionalSenseCodeModeParametersChanged = 0x2A00,
    ScsiAdditionalSenseCodeModeParametersChangedAlias = 0x2A01,
    ScsiAdditionalSenseCodeLogParametersChanged = 0x2A02,
    ScsiAdditionalSenseCodeCommandSequenceError = 0x2C00,
    ScsiAdditionalSenseCodeTaggedCommandsClearedByAnotherInitiator = 0x2F00,
    ScsiAdditionalSenseCodeMediumFormatCorrupted = 0x3100,
    ScsiAdditionalSenseCodeFormatFailed = 0x3101,
    ScsiAdditionalSenseCodeFormatCorruptedBecauseWorldWideNameIsInvalid = 0x3191,
    ScsiAdditionalSenseCodeNoDefectSpareLocationAvailable = 0x3200,
    ScsiAdditionalSenseCodeDefectListUpdateError = 0x3201,
    ScsiAdditionalSenseCodeNoSparesAvailableBecauseTooManyDefectsInOneTrack = 0x3202,
    ScsiAdditionalSenseCodeParameterRounded = 0x3700,
    ScsiAdditionalSenseCodeInvalidBitsInIdentifyMessage = 0x3D00,
    ScsiAdditionalSenseCodeLogicalUnitHasNotSelfConfiguredYet = 0x3E00,
    ScsiAdditionalSenseCodeTargetOperatingConditionsHaveChanged = 0x3F00,
    ScsiAdditionalSenseCodeMicrocodeChanged = 0x3F01,
    ScsiAdditionalSenseCodeChangedOperatingDefinition = 0x3F02,
    ScsiAdditionalSenseCodeInvalidApmParameters = 0x3F90,
    ScsiAdditionalSenseCodeWorldWideNameMismatch = 0x3F91,
    ScsiAdditionalSenseCodeDramParityError = 0x4001,
    ScsiAdditionalSenseCodePowerOnOrSelfTestFailure = 0x4200,
    ScsiAdditionalSenseCodeMessageError = 0x4300,
    ScsiAdditionalSenseCodeInternalTargetError = 0x4400,
    ScsiAdditionalSenseCodeSelectionOrReselectionFailure = 0x4500,
    ScsiAdditionalSenseCodeScsiParityError = 0x4700,
    ScsiAdditionalSenseCodeInitiatorDetectedErrorMessageReceived = 0x4800,
    ScsiAdditionalSenseCodeInvalidMessageError = 0x4900,
    ScsiAdditionalSenseCodeLogicalUnitFailedSelfConfiguration = 0x4C00,
    ScsiAdditionalSenseCodeOverlappedCommandsAttempted = 0x4E00,
    ScsiAdditionalSenseCodeLogException = 0x5B00,                 // factory-installed option
    ScsiAdditionalSenseCodeThresholdConditionMet = 0x5B01,        // factory-installed option
    ScsiAdditionalSenseCodeLogCounterAtMaximum = 0x5B02,          // factory-installed option
    ScsiAdditionalSenseCodeLogListCodesExhausted = 0x5B03,        // factory-installed option
    ScsiAdditionalSenseCodeRplStatusChange = 0x5C00,
    ScsiAdditionalSenseCodeSpindlesSynchronized = 0x5C01,
    ScsiAdditionalSenseCodeSpindlesNotSynchronized = 0x5C02,
    ScsiAdditionalSenseCodeFailurePredictionThresholdExceeded = 0x5D00,
    ScsiAdditionalSenseCodeVoltageFault = 0x6500,
    ScsiAdditionalSenseCodeGeneralFirmwareErrorQualifier = 0x8000,
    ScsiAdditionalSenseCodeReassignPowerFailRecoveryFailed = 0x8100
} ScsiAdditionalSenseCode;

// valid when ScsiSense.SenseKey == ScsiSenseKeyIllegalRequest
//
typedef struct
{
    Bit8 BitPointer : 3;                        // ms bit of field in error
    Bit8 BitPointerIsValid : 1;
    Bit8 Reserved : 2;
    Bit8 CommandNotData : 1;                    // if TRUE, illegal parm was in CDB, else was in data parms sent by initiator during DATA OUT phase
    Bit8 SenseKeySpecificIsValid : 1;
    Bit16 FieldPointer;                         // big-endian order, says which byte of CDB or parm was in error
} ScsiIllegalRequestDetails;

// valid when ScsiSense.SenseKey == ScsiSenseKeyRecoveredError
//    or when ScsiSense.SenseKey == ScsiSenseKeyHardwareError
//    or when ScsiSense.SenseKey == ScsiSenseKeyMediumError
//
typedef struct
{
    Bit8 Reserved : 7;
    Bit8 SenseKeySpecificIsValid : 1;
    Bit16 ActualRetryCount;                     // big-endian order, says how many retries were attempted
} ScsiRecoveredOrHardwareOrMediumErrorDetails;

typedef struct
{
    Bit8 ErrorCode : 7;                         // see enum ScsiErrorCode
    Bit8 Valid : 1;                             // if TRUE, LogicalBlockAddress is valid
    Bit8 SegmentNumber;                         // always 0
    Bit8 SenseKey : 4;                          // see enum ScsiSendKey
    Bit8 ReservedInByte2 : 1;
    Bit8 IncorrectLengthIndication : 1;         // (ILI) previous command's requested block didn't match logical block length on media
    Bit8 EndOfMedium : 1;                       // always 0 for drive
    Bit8 Filemark : 1;                          // always 0 for drive
    Bit32 LogicalBlockAddress;                  // in big-endian byte order, associated with the sense key
    Bit8 AdditionalSenseLength;                 // max is 10 decimal
    Bit8 CommandSpecificData[ 4 ];
    Bit16 AdditionalSenseCode;                  // big-endian, see enum ScsiAdditionalSenseCode
    Bit8 ReservedForSeagateInternalUse;
    union
    {
        ScsiIllegalRequestDetails IllegalRequest;
        ScsiRecoveredOrHardwareOrMediumErrorDetails RecoveredOrHardwareOrMediumError;
    } SenseKeySpecific;
    Bit8 Unused[ 237 ];
} ScsiSense;

typedef enum
{
    ScsiModePageControlCurrent = 0,
    ScsiModePageControlChangeable = 1,
    ScsiModePageControlDefault = 2,
    ScsiModePageControlSaved = 3
} ScsiModePageControl;

typedef enum
{
    ScsiModePageUnitAttention = 0x00,
    ScsiModePageErrorRecovery = 0x01,
    ScsiModePageDisconnnectReconnect = 0x02,
    ScsiModePageFormat = 0x03,
    ScsiModePageRigidDriveGeometry = 0x04,
    ScsiModePageVerifyErrorRecovery = 0x07,
    ScsiModePageCaching = 0x08,
    ScsiModePageControlMode = 0x0A,
    ScsiModePageControlModeAlias = 0x1A,
    ScsiModePageNotchAndPartition = 0x0C,
    ScsiModePagePowerCondition = 0x0D,
    ScsiModePageXor = 0x10,
    ScsiModePageInformationExceptions = 0x1C,
    ScsiModePageAllSupported = 0x3F
} ScsiModePage;

typedef struct
{
    Bit16 SenseDataLength;                      // big-endian
                                                // must be 0 for mode select (10)
    Bit8 MediumType;                            // always zero for ST34501W
    Bit8 Reserved1 : 4;
    Bit8 DpoFuaSupported : 1;                   // if TRUE, DisablePageOut and ForceUnitAccess bits of read and write commands are supported
                                                // must be FALSE for mode select (10)
    Bit8 Reserved2 : 2;
    Bit8 WriteProtected : 1;                    // if TRUE, medium cannot be written to ???but it can save mode settings???
                                                // must be FALSE for mode select (10)
    Bit8 Reserved3[ 2 ];
    Bit16 BlockDescriptorLength;                // Seagate drives return either 8 (one descriptor sent by drive (sense) or host (select)) or 0 (none sent)
} ScsiMode10ParmList;                           // for mode sense (10) and mode select (10)

typedef struct
{
    Bit8 DensityCode;
    Bit8 NumberOfBlocks[ 3 ];
    Bit8 Reserved1;
    Bit8 BlockLength[ 3 ];
} ScsiModeBlockDescriptor;                      // for mode sense (6), mode sense (10), mode select (6), mode select (10)

typedef struct
{
    Bit8 PageCode   : 6;
    Bit8 Reserved1  : 1;
    Bit8 ParameterIsSavable : 1;                // must be FALSE for mode select (6) and (10)
    Bit8 PageLength;
    Bit8 ReadCacheDisable : 1;
    Bit8 MultiplicationFactor : 1;              // if TRUE, min/max prefetch are multiplied by some constant of ???unknown origin
    Bit8 WriteCacheEnable : 1;
    Bit8 SizeEnable : 1;                        // always FALSE for ST34501W, so NumberOfCacheSegments controls size of segments, not CacheSegmentSize
    Bit8 PrefetchAcrossTimeDiscontinuities : 1;
    Bit8 CachingAnalysisPermitted : 1;
    Bit8 AbortPrefetchUponSelection : 1;
    Bit8 InitiatorControl : 1;                  // if TRUE on Seagate drives, disables adaptive read look ahead (ARLA)
    Bit8 DemandReadRetentionPriority : 4;       // always 0 on ST34501W because doesn't distinguish caching host-requested from prefetched data
    Bit8 WriteRetentionPriority : 4;            // always 0 on ST34501W because does distinguish caching host-requested from prefetched data
    Bit16 DisablePrefetchTransferLength;        // big-endian, no prefetch if more than this many blocks in single request
    Bit16 MinimumPrefetch;                      // big-endian, mandatory to prefetch this many sectors ??or is it blocks??
    Bit16 MaximumPrefetch;                      // big-endian, can get up to this many, but may abort beforehand due to ???
    Bit16 MaximumPrefetchCeiling;               // big-endian, drive truncates MaximumPrefetch to this value
    Bit8 Reserved2 : 5;
    Bit8 DisableReadAhead : 1;
    Bit8 LogicalBlockCacheSegmentSize : 1;      // not used by ST34501W today
    Bit8 ForceSequentialWrite : 1;              // if FALSE, target can reorder media writes for command completion speed
    Bit8 NumberOfCacheSegments;
    Bit16 CacheSegmentSize;
    Bit8 Reserved3;
    Bit8 NonCacheSegmentSize[3];                // bytes to reserve for SCSI buffering
} ScsiModeCaching;                              // page 8 from ScsiModeSense



// various CDB's
//

typedef struct
{
    Bit8 Command;                               // 0 for this command
    Bit8 Reserved1 : 5;                         // must be 0
    Bit8 Lun : 3;                               // must be 0
    Bit8 Reserved2[ 3 ];
    Bit8 Link : 1;                              // factory-installed option
    Bit8 Flag : 1;                              // factory-installed option
    Bit8 ReservedControl : 6;
} ScsiTestUnitReady;

typedef struct
{
    Bit8 Command;                               // 3 for this command
    Bit8 Reserved1 : 5;                         // must be 0
    Bit8 Lun : 3;                               // must be 0
    Bit8 Reserved2[ 2 ];
    Bit8 AllocationLength;
    Bit8 Link : 1;
    Bit8 Flag : 1;
    Bit8 ReservedControl : 6;
} ScsiRequestSense;

// I still need to key in lots of data structures
// to match Seagate pages 89-99
//
// can get vendor names, device name, serial number,
// firmware versions, behaviors
//
typedef struct
{
    Bit8 Command;                               // 0x12 for this command
    Bit8 EnableVitalProductData : 1;            // if TRUE, returns vital product data instead of normal inquiry data
    Bit8 Reserved1 : 4;                         // must be 0
    Bit8 Lun : 3;                               // must be 0
    Bit8 PageCode;                              // must be 0 if EnableVitalProductData is FALSE
    Bit8 Reserved2;                             // must be 0
    Bit8 AllocationLength;
    Bit8 Link : 1;
    Bit8 Flag : 1;
    Bit8 ReservedControl : 6;
} ScsiInquiry;

// for returned structure see page 101
//
typedef struct
{
    Bit8 Command;                               // 0x12 for this command
    Bit8 Reserved1 : 5;                         // must be 0
    Bit8 Lun : 3;                               // must be 0
    Bit8 Reserved2;                             // must be 0
    Bit16 AllocationLength;                     // big-endian
    Bit8 Link : 1;
    Bit8 Flag : 1;
    Bit8 ReservedControl : 6;
} ScsiReceiveDiagnosticResults;

// for sent structure see pages 106-107
//
typedef struct
{
    Bit8 Command;                               // 0x1D for this command
    Bit8 UnitOffline : 1;                       // enables touching media or moving heads; drive ignores this
    Bit8 DeviceOffline: 1;                      // enables diags that adversely affect other logical units; ignored by drive
    Bit8 SelfTest : 1;                          // overrides other parms and just does default self-test (seek to reserved cylinder and read/write/verify)
    Bit8 Lun : 3;                               // must be 0
    Bit8 Reserved2;                             // must be 0
    Bit16 ParmListLength;                       // big-endian
    Bit8 Link : 1;
    Bit8 Flag : 1;
    Bit8 ReservedControl : 6;
} ScsiSendDiagnostic;

// note: linked commands are a factory-installed extra-cost option
//

typedef struct
{
    Bit8 Command;                               // 0x28 for this command
    Bit8 AddressIsRelative : 1;                 // if TRUE, LogicalBlockAddress is signed relative to previous CDB which linked to this one
    Bit8 Reserved1 : 2;
    Bit8 ForceUnitAccess : 1;                   // if TRUE, mustn't use cache, and may need to flush cache blocks before reading medium
    Bit8 DisablePageOut : 1;                    // if TRUE, host doesn't think these blocks will be accessed again, so don't bother to cache them
    Bit8 Lun : 3;                               // must be 0
    Bit32 LogicalBlockAddress;                  // big-endian
    Bit8 Reserved2;                             // must be 0
    Bit16 TransferLength;                       // big-endian, in blocks, 0 means none
    Bit8 Link : 1;                              // if TRUE, disk returns INTERMEDIATE STATUS and sends msg defined by flag field
    Bit8 Flag : 1;                              // if Link and me TRUE, sends LINKED COMMAND COMPLETE (WITH FLAG), otherwise
    Bit8 ReservedControl : 6;
} ScsiReadExtended;

typedef struct
{
    Bit8 Command;                               // 0x2A for this command
    Bit8 AddressIsRelative : 1;                 // if TRUE, LogicalBlockAddress is signed relative to previous CDB which linked to this one
    Bit8 Reserved1 : 2;
    Bit8 ForceUnitAccess : 1;                   // if TRUE, mustn't return GOOD until data written to medium
    Bit8 DisablePageOut : 1;                    // if TRUE, host doesn't think these blocks will be accessed again, so don't bother to cache them
    Bit8 Lun : 3;                               // must be 0
    Bit32 LogicalBlockAddress;                  // big-endian
    Bit8 Reserved2;                             // must be 0
    Bit16 TransferLength;                       // big-endian, in blocks, 0 means none
    Bit8 Link : 1;                              // if TRUE, disk returns INTERMEDIATE STATUS and sends msg defined by flag field
    Bit8 Flag : 1;                              // if Link and me TRUE, sends LINKED COMMAND COMPLETE (WITH FLAG), otherwise
    Bit8 ReservedControl : 6;                   // must be 0
} ScsiWriteExtended;

// transfers one logical block worth of data from initiator
// to drive, but writes it to many blocks on the disk
//
// note: LogicalBlockData and PhysicalBlockData cannot both be TRUE
//
typedef struct
{
    Bit8 Command;                               // 0x41 for this command
    Bit8 AddressIsRelative : 1;                 // if TRUE, LogicalBlockAddress is signed relative to previous CDB which linked to this one
    Bit8 LogicalBlockData : 1;                  // if TRUE, replace first 4 bytes of each block with its LBA (fails with no sense details!)
    Bit8 PhysicalBlockData : 1;                 // if TRUE, replace first 8 bytes of each sector with its physical address (fails with no sense details!)
    Bit8 Reserved1 : 2;
    Bit8 Lun : 3;                               // must be 0
    Bit32 LogicalBlockAddress;                  // big-endian
    Bit8 Reserved2;                             // must be 0
    Bit16 NumberOfBlocks;                       // big-endian, 0 means remainder of medium
    Bit8 ReservedControl;                       // must be 0
} ScsiWriteSame;

// gets various drive parameters
//
typedef struct
{
    Bit8 Command;                               // 0x5A for this command
    Bit8 Reserved1 : 3;
    Bit8 DisableBlockDescriptors : 1;           // if TRUE, no ScsiModeBlockDescriptor will be returned
    Bit8 Reserved2 : 1;
    Bit8 Lun : 3;                               // must be 0
    Bit8 PageCode : 5;                          // see enum ScsiModePage
    Bit8 PageControl : 2;                       // see enum ScsiModePageControl
    Bit8 Reserved3[ 4 ];
    Bit16 AllocationLength;                     // big-endian
    Bit8 ReservedControl;                       // must be 0
} ScsiModeSense10;

// sets various drive parameters
//
typedef struct
{
    Bit8 Command;                               // 0x55 for this command
    Bit8 SaveModeParameters : 1;                // if TRUE, parms are written to media before returning GOOD status; if FALSE, only current parms are affected
    Bit8 Reserved1 : 3;
    Bit8 PageFormat : 1;                        // assumed by drive to be TRUE (FALSE would mean data after block descriptors was in vendor-specific format)
    Bit8 Lun : 3;                               // must be 0
    Bit8 Reserved2[ 5 ];
    Bit16 ParmListLength;                       // big-endian
    Bit8 ReservedControl;                       // must be 0
} ScsiModeSelect10;



typedef union
{
    ScsiTestUnitReady TestUnitReady;
    ScsiRequestSense RequestSense;
    ScsiInquiry Inquiry;
    ScsiReceiveDiagnosticResults ReceiveDiagnosticResults;
    ScsiSendDiagnostic SendDiagnostic;
    ScsiReadExtended ReadExtended;
    ScsiWriteExtended WriteExtended;
    ScsiWriteSame WriteSame;
    ScsiModeSense10 ModeSense10;
    ScsiModeSelect10 ModeSelect10;
} ScsiCdb;



#ifdef __cplusplus
}
#endif



#endif      // not included yet
